#
# Mixins.py -- Mixin classes for FITS viewer.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

from ginga.misc.Callback import Callbacks


class UIMixin(object):

    def __init__(self):
        self.ui_active = False
        self.ui_viewer = set([])

        for name in ('motion', 'button-press', 'button-release',
                     'key-press', 'key-release', 'drag-drop',
                     'scroll', 'map', 'focus', 'enter', 'leave',
                     'cursor-changed'):
            self.enable_callback(name)

    def ui_is_active(self):
        return self.ui_active

    def ui_set_active(self, tf, viewer=None):
        self.ui_active = tf
        if viewer is not None:
            if tf:
                self.ui_viewer.add(viewer)
            else:
                if viewer in self.ui_viewer:
                    self.ui_viewer.remove(viewer)

    def make_ui_callback(self, name, *args, **kwargs):
        """Invoke callbacks on all objects (i.e. layers) from the top to
        the bottom, returning when the first one returns True.  If none
        returns True, then make the callback on our 'native' layer.
        """
        if hasattr(self, 'objects'):
            # Invoke callbacks on all our layers that have the UI mixin
            num = len(self.objects) - 1
            while num >= 0:
                obj = self.objects[num]
                if isinstance(obj, UIMixin) and obj.ui_is_active():
                    res = obj.make_ui_callback(name, *args, **kwargs)
                    if res:
                        return res
                num -= 1

        if self.ui_active:
            return super(UIMixin, self).make_callback(name, *args, **kwargs)

        return False

    def make_ui_callback_viewer(self, viewer, name, *args, **kwargs):
        """Invoke callbacks on all objects (i.e. layers) from the top to
        the bottom, returning when the first one returns True.  If none
        returns True, then make the callback on our 'native' layer.
        """
        if len(self.ui_viewer) == 0 or viewer in self.ui_viewer:
            if hasattr(self, 'objects'):
                # Invoke callbacks on all our layers that have the UI mixin
                num = len(self.objects) - 1
                while num >= 0:
                    obj = self.objects[num]
                    if isinstance(obj, UIMixin) and obj.ui_is_active():
                        res = obj.make_ui_callback_viewer(viewer, name,
                                                          *args, **kwargs)
                        if res:
                            return res
                    num -= 1

            if self.ui_active:
                return super(UIMixin, self).make_callback(name, *args, **kwargs)

        return False

    def make_callback_children(self, name, *args, **kwargs):
        """Invoke callbacks on all objects (i.e. layers) from the top to
        the bottom, returning when the first one returns True.  If none
        returns True, then make the callback on our 'native' layer.
        """
        if hasattr(self, 'objects'):
            # Invoke callbacks on all our layers that have the UI mixin
            num = len(self.objects) - 1
            while num >= 0:
                obj = self.objects[num]
                if isinstance(obj, Callbacks):
                    res = obj.make_callback(name, *args, **kwargs)
                    if res:
                        return res
                num -= 1

        return super(UIMixin, self).make_callback(name, *args, **kwargs)



#
# ImageViewGtk.py -- a backend for Ginga using Gtk widgets and Cairo
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import sys
import numpy as np

from ginga.gtk4w import GtkHelp
from ginga import ImageView, Mixins, Bindings
from ginga.cursors import cursor_info
from ginga.canvas import render

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf
import cairo

have_opengl = False
try:
    from ginga.opengl.GlHelp import get_transforms
    from ginga.opengl.glsl import req
    have_opengl = True
except ImportError:
    pass


class ImageViewGtkError(ImageView.ImageViewError):
    pass


class ImageViewGtk(ImageView.ImageViewBase):

    def __init__(self, logger=None, rgbmap=None, settings=None, render=None):
        ImageView.ImageViewBase.__init__(self, logger=logger,
                                         rgbmap=rgbmap,
                                         settings=settings)

        if render is None:
            render = self.t_.get('render_widget', 'widget')
        self.wtype = render
        self.surface = None
        if self.wtype == 'widget':
            imgwin = Gtk.DrawingArea()

            imgwin.set_draw_func(self.draw_event, None)
            imgwin.connect("resize", self.configure_event)

            renderers = ['cairo', 'agg', 'pil', 'opencv']
            self.t_.set_defaults(renderer='cairo')
            if self.t_['renderer'] == 'opengl':
                # currently cannot use opengl renderer except with GLArea
                self.t_.set(renderer='cairo')

            if sys.byteorder == 'little':
                self.rgb_order = 'BGRA'
            else:
                self.rgb_order = 'RGBA'

        elif self.wtype == 'opengl':
            if not have_opengl:
                raise ImageViewGtkError("Please install 'pyopengl' to use render: '%s'" % (render))
            # NOTE: See https://gitlab.gnome.org/GNOME/gtk/issues/1270 for
            # an issue regarding a buggy GLX/Mesa driver for X11 on Linux;
            # if you experience non-GL widgets flashing when using the
            # opengl renderer then try setting the following environment
            # variable:
            #   GDK_GL=software-draw-surface
            #
            imgwin = Gtk.GLArea()
            imgwin.set_required_version(req.major, req.minor)
            imgwin.set_has_depth_buffer(False)
            imgwin.set_has_stencil_buffer(False)
            imgwin.set_auto_render(False)

            imgwin.connect('realize', self.on_realize_cb)
            imgwin.connect('render', self.on_render_cb)
            imgwin.connect("resize", self.configure_glarea_cb)

            renderers = ['opengl']
            # currently can only use opengl renderer with GLArea
            #self.t_.set_defaults(renderer='opengl')
            self.t_.set(renderer='opengl')
            self.rgb_order = 'RGBA'

            # we replace some transforms in the catalog for OpenGL rendering
            self.tform = get_transforms(self)

        else:
            raise ImageViewGtkError("Undefined render type: '%s'" % (render))

        imgwin.set_hexpand(True)
        imgwin.set_vexpand(True)
        self.imgwin = imgwin

        # see reschedule_redraw() method
        self._defer_task = GtkHelp.Timer()
        self._defer_task.add_callback('expired',
                                      lambda timer: self.delayed_redraw())
        self.msgtask = GtkHelp.Timer()
        self.msgtask.add_callback('expired',
                                  lambda timer: self.onscreen_message(None))

        self.renderer = None
        # Pick a renderer that can work with us
        preferred = self.t_['renderer']
        if preferred in renderers:
            renderers.remove(preferred)
        self.possible_renderers = [preferred] + renderers
        self.choose_best_renderer()

    def get_widget(self):
        return self.imgwin

    def choose_renderer(self, name):
        if self.wtype == 'opengl':
            if name != 'opengl':
                raise ValueError("Only possible renderer for this widget "
                                 "is 'opengl'")
        klass = render.get_render_class(name)
        self.renderer = klass(self)

        wd = max(1, self.imgwin.get_width())
        ht = max(1, self.imgwin.get_height())
        self.configure_window(wd, ht)

    def choose_best_renderer(self):
        for name in self.possible_renderers:
            try:
                self.choose_renderer(name)
                self.logger.info("best renderer available is '{}'".format(name))
                return
            except Exception as e:
                # uncomment to troubleshoot
                self.logger.error("Error choosing renderer '{}': {}".format(name, e),
                                  exc_info=True)
                continue

        raise ImageViewGtkError("No valid renderers available: {}".format(str(self.possible_renderers)))

    def get_plain_image_as_pixbuf(self):
        arr = self.getwin_array(order='RGB', dtype=np.uint8)
        pixbuf = GtkHelp.pixbuf_new_from_array(arr,
                                               GdkPixbuf.Colorspace.RGB,
                                               8)
        return pixbuf

    def get_plain_image_as_widget(self):
        """Returns a Gtk.Image widget of the images displayed.
        Does not include overlaid graphics.
        """
        pixbuf = self.get_plain_image_as_pixbuf()
        image = Gtk.Image()
        image.set_from_pixbuf(pixbuf)
        image.show()
        return image

    def save_plain_image_as_file(self, filepath, format='png', quality=90):
        """Does not include overlaid graphics."""
        pixbuf = self.get_plain_image_as_pixbuf()
        options, values = [], []
        if format == 'jpeg':
            options.append('quality')
            values.append(str(quality))
        pixbuf.savev(filepath, format, options, values)

    def reschedule_redraw(self, time_sec):
        self._defer_task.stop()
        self._defer_task.start(time_sec)

    def _renderer_to_surface(self):

        if isinstance(self.renderer.surface, cairo.ImageSurface):
            # optimization when renderer is cairo:
            # the render already contains a surface we can copy from
            self.surface = self.renderer.surface

        else:
            # create a new surface from rendered array
            arr = self.renderer.get_surface_as_array(order=self.rgb_order)
            arr = np.ascontiguousarray(arr)

            daht, dawd, depth = arr.shape
            if dawd == 0 or daht == 0:
                # can happen if we get a draw event before widget is mapped
                return
            self.logger.debug("arr shape is %dx%dx%d" % (dawd, daht, depth))

            stride = cairo.ImageSurface.format_stride_for_width(cairo.FORMAT_ARGB32,
                                                                dawd)
            self.surface = cairo.ImageSurface.create_for_data(arr,
                                                              cairo.FORMAT_ARGB32,
                                                              dawd, daht, stride)

    def update_widget(self):
        if self.imgwin is None:
            return

        self.logger.debug("updating window")
        if self.wtype == 'opengl':
            self.imgwin.queue_render()
            return

        self._renderer_to_surface()

        has_window = self.imgwin.get_mapped()
        if has_window and self.surface is not None:
            imgwin_wd, imgwin_ht = self.get_window_size()

            self.imgwin.queue_draw()

    def draw_event(self, widget, cr, width, height, user_data):
        if self.surface is None:
            # window is not mapped/configured yet
            return
        #self.logger.debug("updating window from surface")
        # redraw the screen from backing surface
        cr.set_source_surface(self.surface, 0, 0)

        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.paint()
        return False

    def configure_window(self, width, height):
        self.logger.debug("window size reconfigured to %dx%d" % (
            width, height))
        self.configure(width, height)

    def configure_event(self, widget, width, height):
        # NOTE: this callback is only for the DrawingArea widget
        if self.surface is not None:
            # This is a workaround for a issue in Gtk
            # where we get multiple configure callbacks even though
            # the size hasn't changed.  We avoid creating a new surface
            # if there is an old surface with the exact same size.
            # This prevents some flickering of the display on focus events.
            wwd, wht = self.get_window_size()
            if (wwd == width) and (wht == height):
                return True

        self.logger.debug(f"allocation is {width}x{height}")
        self.configure_window(width, height)
        return True

    def configure_glarea_cb(self, widget, width, height):
        # NOTE: this callback is only for the GLArea (OpenGL) widget
        self.logger.debug("allocation is %dx%d" % (width, height))
        self.configure_window(width, height)
        return True

    def make_context_current(self):
        ctx = self.imgwin.get_context()
        if ctx is not None:
            ctx.make_current()
        return ctx

    def on_realize_cb(self, area):
        # NOTE: this callback is only for the GLArea (OpenGL) widget
        self.renderer.gl_initialize()

    def on_render_cb(self, area, ctx):
        # NOTE: this callback is only for the GLArea (OpenGL) widget
        self.renderer.gl_paint()

    def prepare_image(self, cvs_img, cache, whence):
        self.renderer.prepare_image(cvs_img, cache, whence)

    def set_cursor(self, cursor):
        self.imgwin.set_cursor(cursor)

    def make_cursor(self, iconpath, x, y, size=None):
        if size is None:
            def_px_size = self.settings.get('default_cursor_length', 16)
            size = (def_px_size, def_px_size)
        cursor = GtkHelp.make_cursor(self.imgwin, iconpath, x, y, size=size)
        return cursor

    def center_cursor(self):
        # NOTE: disp.warp_pointer() not supported in Gtk4
        pass

    def position_cursor(self, data_x, data_y):
        # NOTE: disp.warp_pointer() not supported in Gtk4
        pass

    def make_timer(self):
        return GtkHelp.Timer()

    def onscreen_message(self, text, delay=None, redraw=True):
        self.msgtask.stop()
        self.set_onscreen_message(text, redraw=redraw)
        if delay is not None:
            self.msgtask.start(delay)

    def take_focus(self):
        self.imgwin.grab_focus()


class GtkEventMixin(object):

    def __init__(self):
        imgwin = self.imgwin
        imgwin.set_can_focus(True)
        imgwin.set_focusable(True)
        #imgwin.set_focus_on_click(True)
        imgwin.connect("realize", self.map_event)
        event = Gtk.EventControllerFocus.new()
        event.connect("enter", self.focus_event, imgwin, True)
        imgwin.add_controller(event)
        event = Gtk.EventControllerFocus.new()
        event.connect("leave", self.focus_event, imgwin, False)
        imgwin.add_controller(event)
        event = Gtk.EventControllerMotion.new()
        event.connect("enter", self.enter_notify_event, imgwin)
        imgwin.add_controller(event)
        event = Gtk.EventControllerMotion.new()
        event.connect("leave", self.leave_notify_event, imgwin)
        imgwin.add_controller(event)
        event = Gtk.EventControllerMotion.new()
        event.connect("motion", self.motion_notify_event, imgwin)
        imgwin.add_controller(event)
        event = Gtk.GestureClick.new()
        event.connect("pressed", self.button_press_event, imgwin)
        # all buttons
        event.set_button(0)
        imgwin.add_controller(event)
        event = Gtk.GestureClick.new()
        event.connect("released", self.button_release_event, imgwin)
        event.set_button(0)
        imgwin.add_controller(event)
        event = Gtk.EventControllerKey.new()
        event.connect("key-pressed", self.key_press_event, imgwin)
        imgwin.add_controller(event)
        event = Gtk.EventControllerKey.new()
        event.connect("key-released", self.key_release_event, imgwin)
        imgwin.add_controller(event)
        event = Gtk.EventControllerScroll.new(
            Gtk.EventControllerScrollFlags.BOTH_AXES |
            Gtk.EventControllerScrollFlags.KINETIC)
        event.connect("scroll", self.scroll_event, imgwin)
        imgwin.add_controller(event)

        # Set up widget as a drag and drop destination
        drop_tgt = Gtk.DropTarget.new(Gdk.FileList, Gdk.DragAction.COPY)
        drop_tgt.connect('drop', self.on_dnd_drop)
        drop_tgt.connect('accept', self.on_dnd_accept)
        drop_tgt.connect('enter', self.on_dnd_enter)
        drop_tgt.connect('motion', self.on_dnd_motion)
        drop_tgt.connect('leave', self.on_dnd_leave)
        imgwin.add_controller(drop_tgt)

        # for gtk key handling
        self._keytbl = {
            'shift_l': 'shift_l',
            'shift_r': 'shift_r',
            'control_l': 'control_l',
            'control_r': 'control_r',
            'alt_l': 'alt_l',
            'alt_r': 'alt_r',
            'super_l': 'super_l',
            'super_r': 'super_r',
            'meta_right': 'meta_right',
            'asciitilde': '~',
            'grave': 'backquote',
            'exclam': '!',
            'at': '@',
            'numbersign': '#',
            'percent': '%',
            'asciicircum': '^',
            'ampersand': '&',
            'asterisk': '*',
            'dollar': '$',
            'parenleft': '(',
            'parenright': ')',
            'underscore': '_',
            'minus': '-',
            'plus': '+',
            'equal': '=',
            'braceleft': '{',
            'braceright': '}',
            'bracketleft': '[',
            'bracketright': ']',
            'bar': '|',
            'colon': ':',
            'semicolon': ';',
            'quotedbl': 'doublequote',
            'apostrophe': 'singlequote',
            'backslash': 'backslash',
            'less': '<',
            'greater': '>',
            'comma': ',',
            'period': '.',
            'question': '?',
            'slash': '/',
            'space': 'space',
            'escape': 'escape',
            'return': 'return',
            'tab': 'tab',
            'f1': 'f1',
            'f2': 'f2',
            'f3': 'f3',
            'f4': 'f4',
            'f5': 'f5',
            'f6': 'f6',
            'f7': 'f7',
            'f8': 'f8',
            'f9': 'f9',
            'f10': 'f10',
            'f11': 'f11',
            'f12': 'f12',
            'right': 'right',
            'left': 'left',
            'up': 'up',
            'down': 'down',
            'insert': 'insert',
            'delete': 'delete',
            'home': 'home',
            'end': 'end',
            'page_up': 'page_up',
            'page_down': 'page_down',
        }

        # Define cursors
        cursor_names = cursor_info.get_cursor_names()
        def_px_size = self.settings.get('default_cursor_length', 16)
        for curname in cursor_names:
            curinfo = cursor_info.get_cursor_info(curname)
            wd_px = int(curinfo.scale_width * def_px_size)
            ht_px = int(curinfo.scale_height * def_px_size)
            pt_x = int(curinfo.point_x_pct * wd_px)
            pt_y = int(curinfo.point_y_pct * ht_px)
            cur = self.make_cursor(curinfo.path, pt_x, pt_y,
                                   size=(wd_px, ht_px))
            self.define_cursor(curinfo.name, cur)

        for name in ('motion', 'button-press', 'button-release',
                     'key-press', 'key-release', 'drag-drop',
                     'scroll', 'map', 'focus', 'enter', 'leave',
                     ):
            self.enable_callback(name)

    def transkey(self, keyname):
        try:
            return self._keytbl[keyname.lower()]

        except KeyError:
            return keyname

    def get_key_table(self):
        return self._keytbl

    def map_event(self, widget):
        self.switch_cursor('pick')

        return self.make_callback('map')

    def focus_event(self, controller, widget, hasFocus):
        return self.make_callback('focus', hasFocus)

    def enter_notify_event(self, controller, x, y, widget):
        self.last_win_x, self.last_win_y = x, y

        self.check_cursor_location()

        enter_focus = self.t_.get('enter_focus', False)
        if enter_focus:
            widget.grab_focus()
        return self.make_callback('enter')

    def leave_notify_event(self, controller, widget):
        self.logger.debug("leaving widget...")
        return self.make_callback('leave')

    def key_press_event(self, controller, keyval, keycode, state, widget):
        keyname = Gdk.keyval_name(keyval)
        keyname = self.transkey(keyname)
        self.logger.debug("key press event, key=%s" % (keyname))
        return self.make_ui_callback_viewer(self, 'key-press', keyname)

    def key_release_event(self, controller, keyval, keycode, state, widget):
        keyname = Gdk.keyval_name(keyval)
        keyname = self.transkey(keyname)
        self.logger.debug("key release event, key=%s" % (keyname))
        return self.make_ui_callback_viewer(self, 'key-release', keyname)

    def button_press_event(self, gesture, whatsit, x, y, widget):
        self.last_win_x, self.last_win_y = x, y
        event_button = gesture.get_current_button()
        button = 0
        if event_button != 0:
            button |= 0x1 << (event_button - 1)
        self.logger.debug("button event at %dx%d, button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        return self.make_ui_callback_viewer(self, 'button-press', button,
                                            data_x, data_y)

    def button_release_event(self, gesture, whatsit, x, y, widget):
        self.last_win_x, self.last_win_y = x, y

        event_button = gesture.get_current_button()
        button = 0
        if event_button != 0:
            button |= 0x1 << (event_button - 1)
        self.logger.debug("button release at %dx%d button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        return self.make_ui_callback_viewer(self, 'button-release', button,
                                            data_x, data_y)

    def motion_notify_event(self, motion, x, y, widget):
        button = 0
        self.last_win_x, self.last_win_y = x, y

        state = motion.get_current_event_state()
        if state & Gdk.ModifierType.BUTTON1_MASK:
            button |= 0x1
        elif state & Gdk.ModifierType.BUTTON2_MASK:
            button |= 0x2
        elif state & Gdk.ModifierType.BUTTON3_MASK:
            button |= 0x4
        self.logger.debug("motion event at %dx%d, button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback_viewer(self, 'motion', button,
                                            data_x, data_y)

    def scroll_event(self, controller, dx, dy, widget):
        degrees, direction = GtkHelp.get_scroll_info(dx, dy)
        self.logger.debug("scroll deg=%f direction=%f" % (
            degrees, direction))

        # pointer location should have been recorded by motion_notify_event
        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback_viewer(self, 'scroll', direction, degrees,
                                            data_x, data_y)

    def on_dnd_drop(self, drop_target, value, x, y):
        paths = [gdk_path.get_path() for gdk_path in value]
        self.logger.debug("dropped filename(s): %s" % (str(paths)))
        if len(paths) > 0:
            self.make_ui_callback_viewer(self, 'drag-drop', paths)

    def on_dnd_accept(self, drop, user_data):
        # TODO: double-check drop type
        return True

    def on_dnd_enter(self, drop_target, x, y):
        return Gdk.DragAction.COPY

    def on_dnd_motion(self, drop_target, x, y):
        return Gdk.DragAction.COPY

    def on_dnd_leave(self, user_data):
        # For possible future use
        pass


class ImageViewEvent(Mixins.UIMixin, GtkEventMixin, ImageViewGtk):

    def __init__(self, logger=None, rgbmap=None, settings=None, render=None):
        ImageViewGtk.__init__(self, logger=logger, rgbmap=rgbmap,
                              settings=settings, render=render)
        Mixins.UIMixin.__init__(self)
        GtkEventMixin.__init__(self)


class ImageViewZoom(ImageViewEvent):

    # class variables for binding map and bindings can be set
    bindmapClass = Bindings.BindingMapper
    bindingsClass = Bindings.ImageViewBindings

    @classmethod
    def set_bindingsClass(cls, klass):
        cls.bindingsClass = klass

    @classmethod
    def set_bindmapClass(cls, klass):
        cls.bindmapClass = klass

    def __init__(self, logger=None, rgbmap=None, settings=None,
                 render=None, bindmap=None, bindings=None):
        ImageViewEvent.__init__(self, logger=logger, rgbmap=rgbmap,
                                settings=settings, render=render)

        self.ui_set_active(True, viewer=self)

        if bindmap is None:
            bindmap = ImageViewZoom.bindmapClass(self.logger)
        self.bindmap = bindmap
        bindmap.register_for_events(self)

        if bindings is None:
            bindings = ImageViewZoom.bindingsClass(self.logger)
        self.set_bindings(bindings)

    def get_bindmap(self):
        return self.bindmap

    def get_bindings(self):
        return self.bindings

    def set_bindings(self, bindings):
        self.bindings = bindings
        bindings.set_bindings(self)


class CanvasView(ImageViewZoom):
    """A Ginga viewer for viewing 2D slices of image data."""

    def __init__(self, logger=None, settings=None, rgbmap=None,
                 render=None, bindmap=None, bindings=None):
        ImageViewZoom.__init__(self, logger=logger, settings=settings,
                               rgbmap=rgbmap, render=render,
                               bindmap=bindmap, bindings=bindings)

        # Needed for UIMixin to propagate events correctly
        self.objects = [self.private_canvas]

    def set_canvas(self, canvas, private_canvas=None):
        super(CanvasView, self).set_canvas(canvas,
                                           private_canvas=private_canvas)

        self.objects[0] = self.private_canvas


class ScrolledView(Gtk.Grid):
    """A class that can take a viewer as a parameter and add scroll bars
    that respond to the pan/zoom levels.
    """

    def __init__(self, viewer, parent=None):
        self.viewer = viewer
        super(ScrolledView, self).__init__()

        GtkHelp.set_border_width(self, 0)
        self.set_row_spacing(0)
        self.set_column_spacing(0)

        self._adjusting = False
        self._scrolling = False
        self.pad = 20
        self.sb_thickness = 20
        self.rng_x = 100.0
        self.rng_y = 100.0
        self._bar_status = dict(horizontal='on', vertical='on')

        # reparent the viewer widget
        self.v_w = viewer.get_widget()
        self.attach(self.v_w, 0, 0, 1, 1)

        self.hsb = Gtk.Scrollbar(orientation=Gtk.Orientation.HORIZONTAL)
        self.hsb.set_size_request(-1, self.sb_thickness)
        adj = self.hsb.get_adjustment()
        adj.connect('value-changed', self._scroll_contents)
        self.vsb = Gtk.Scrollbar(orientation=Gtk.Orientation.VERTICAL)
        self.vsb.set_size_request(self.sb_thickness, -1)
        adj = self.vsb.get_adjustment()
        adj.connect('value-changed', self._scroll_contents)

        self.viewer.add_callback('redraw', self._calc_scrollbars)
        self.viewer.add_callback('limits-set',
                                 lambda v, l: self._calc_scrollbars(v, 0))

        self._calc_scrollbars(self.viewer, 0)

    def get_widget(self):
        return self

    def _calc_scrollbars(self, viewer, whence):
        """Calculate and set the scrollbar handles from the pan and
        zoom positions.
        """
        if self._scrolling or whence > 0:
            return

        # flag that suppresses a cyclical callback
        self._adjusting = True
        try:
            res = self.viewer.calc_pan_pct(pad=self.pad)
            if res is None:
                return

            page_x, page_y = (float(res.thm_pct_x * 100),
                              float(res.thm_pct_y * 100))
            self.rng_x, self.rng_y = 100 - page_x, 100 - page_y
            val_x, val_y = (float(res.pan_pct_x * self.rng_x),
                            float((1.0 - res.pan_pct_y) * self.rng_y))

            upper_x, upper_y = 100.0, 100.0

            adj = self.hsb.get_adjustment()
            adj.configure(val_x, 0.0, upper_x, 1.0, page_x, page_x)
            self.hsb.set_adjustment(adj)
            adj = self.vsb.get_adjustment()
            adj.configure(val_y, 0.0, upper_y, 1.0, page_y, page_y)
            self.vsb.set_adjustment(adj)

        finally:
            self._adjusting = False
        return True

    def _scroll_contents(self, adj):
        """Called when the scroll bars are adjusted by the user.
        """
        if self._adjusting:
            return True

        self._scrolling = True
        try:
            pos_x = self.hsb.get_adjustment().get_value()
            pos_y = self.vsb.get_adjustment().get_value()

            pct_x = pos_x / float(self.rng_x)
            # invert Y pct because of orientation of scrollbar
            pct_y = 1.0 - (pos_y / float(self.rng_y))

            self.viewer.pan_by_pct(pct_x, pct_y, pad=self.pad)

            # This shouldn't be necessary, but seems to be
            self.viewer.redraw(whence=0)

        finally:
            self._scrolling = False

        return True

    def scroll_bars(self, horizontal='on', vertical='on'):
        # TODO: 'auto' setting working the same as 'on'--fix
        self._bar_status.update(dict(horizontal=horizontal,
                                     vertical=vertical))
        if horizontal in ('on', 'auto'):
            self.attach(self.hsb, 0, 1, 1, 1)
            self.hsb.show()
        elif horizontal == 'off':
            self.remove(self.hsb)
        else:
            raise ValueError("Bad scroll bar option: '%s'; should be one of ('on', 'off' or 'auto')" % (horizontal))

        if vertical in ('on', 'auto'):
            self.attach(self.vsb, 1, 0, 1, 1)
            self.vsb.show()
        elif vertical == 'off':
            self.remove(self.vsb)
        else:
            raise ValueError("Bad scroll bar option: '%s'; should be one of ('on', 'off' or 'auto')" % (vertical))

    def get_scroll_bars_status(self):
        return self._bar_status



#
# Plotable.py -- Abstraction of generic plot data.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga.BaseImage import ViewerObjectBase, Header
from ginga.canvas.CanvasObject import get_canvas_types
from ginga.misc.Bunch import Bunch


class Plotable(ViewerObjectBase):
    """Abstraction of a plot data.

    .. note:: This module is NOT thread-safe!

    """
    # class variables for WCS can be set
    wcsClass = None
    ioClass = None

    @classmethod
    def set_wcsClass(cls, klass):
        cls.wcsClass = klass

    @classmethod
    def set_ioClass(cls, klass):
        cls.ioClass = klass

    def __init__(self, data_ap=None, metadata=None, logger=None, name=None,
                 wcsclass=wcsClass, ioclass=ioClass):

        ViewerObjectBase.__init__(self, logger=logger, metadata=metadata,
                                  name=name)

        self.wcs = None
        self.io = None

        dc = get_canvas_types()
        self.canvas = dc.Canvas()
        self.titles = Bunch(title=None, x_axis=None, y_axis=None)
        self.grid = False
        self.rgb_order = 'RGBA'

    def get_size(self):
        return (self.columns, self.rows)

    def get_canvas(self):
        return self.canvas

    def get_header(self, create=True, include_primary_header=None):
        # By convention, the header is stored in a dictionary
        # under the metadata keyword 'header'
        if 'header' not in self:
            if not create:
                # TODO: change to ValueError("No header found")
                raise KeyError('header')

            hdr = Header()
            self.set(header=hdr)
        else:
            hdr = self['header']

        return hdr

    def has_primary_header(self):
        return False

    def set_titles(self, title=None, x_axis=None, y_axis=None):
        if x_axis is not None:
            self.titles.x_axis = x_axis
        if y_axis is not None:
            self.titles.y_axis = y_axis
        if title is not None:
            self.titles.title = title

    def set_grid(self, tf):
        self.grid = tf

    def clear(self):
        self.canvas.delete_all_objects()
        self.titles.x_axis = None
        self.titles.y_axis = None
        self.titles.title = None
        self.grid = False

    def get_minmax(self, noinf=False):
        # TODO: what should this mean for a plot?
        return (0, 0)

    def load_file(self, filespec):
        raise NotImplementedError("This method is not yet implemented")

    def get_thumbnail(self, length):
        thumb_np = np.eye(length)
        return thumb_np



#
# CanvasRenderPil.py -- for rendering into a Ginga widget with pillow
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import math
import numpy as np

from PIL import Image

from ginga.canvas import render
# force registration of all canvas types
import ginga.canvas.types.all  # noqa
from ginga import trcalc

from . import PilHelp


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)

        # TODO: encapsulate this drawable
        self.ctx = PilHelp.PilContext(surface)

        # special scaling for PIL text drawing to normalize it relative
        # to other backends
        self._font_scale_factor = 1.75

    def get_line(self, color, alpha=1.0, linewidth=1, linestyle='solid'):
        line = super().get_line(color, alpha=alpha, linewidth=linewidth,
                                linestyle=linestyle)

        line.render.color = line.get_bpp_color(8)
        return line

    def get_fill(self, color, alpha=1.0):
        fill = super().get_fill(color, alpha=alpha)

        fill.render.color = fill.get_bpp_color(8)
        return fill

    def get_font(self, fontname, **kwargs):
        font = super().get_font(fontname, **kwargs)

        font.render.font = self.ctx._get_font(font)
        return font

    def text_extents(self, text, font=None):
        if font is None:
            font = self.font
        _font = font.render.font
        if hasattr(_font, 'getbbox'):
            # PIL v10.0
            l, t, r, b = _font.getbbox(text)
            wd_px, ht_px = int(abs(round(r - l))), int(abs(round(b - t)))
        else:
            wd_px, ht_px = _font.getsize(text)
        return wd_px, ht_px

    def get_affine_transform(self, cx, cy, rot_deg):
        x, y = 0, 0          # old center
        nx, ny = cx, cy      # new center
        sx = sy = 1.0        # new scale
        cosine = math.cos(math.radians(rot_deg))
        sine = math.sin(math.radians(rot_deg))
        a = cosine / sx
        b = sine / sx
        c = x - nx * a - ny * b
        d = -sine / sy
        e = cosine / sy
        f = y - nx * d - ny * e
        return (a, b, c, d, e, f)

    ##### DRAWING OPERATIONS #####

    def draw_image(self, cvs_img, cpoints, rgb_arr, whence, order='RGBA'):
        # no-op for this renderer
        pass

    def draw_text(self, cx, cy, text, rot_deg=0.0, font=None, fill=None,
                  line=None):
        wd, ht = self.ctx.text_extents(text, font=font)

        # NOTE: rotation ignored in PIL, for now
        self.ctx.text((cx, cy - ht), text, font, line, fill)

    def draw_polygon(self, cpoints, line=None, fill=None):
        cpoints = trcalc.strip_z(cpoints)
        self.ctx.polygon(cpoints, line, fill)

    def draw_circle(self, cx, cy, cradius, line=None, fill=None):
        self.ctx.circle((cx, cy), cradius, line, fill)

    def draw_line(self, cx1, cy1, cx2, cy2, line=None):
        self.ctx.line((cx1, cy1), (cx2, cy2), line)

    def draw_path(self, cpoints, line=None):
        cpoints = trcalc.strip_z(cpoints)
        self.ctx.path(cpoints, line)


class CanvasRenderer(render.StandardPipelineRenderer):

    def __init__(self, viewer):
        render.StandardPipelineRenderer.__init__(self, viewer)

        self.kind = 'pil'
        self.rgb_order = 'RGBA'
        self.surface = None
        self.dims = ()

    def resize(self, dims):
        """Resize our drawing area to encompass a space defined by the
        given dimensions.
        """
        width, height = dims[:2]
        self.logger.debug("renderer reconfigured to %dx%d" % (
            width, height))

        # create PIL surface the size of the window
        # NOTE: pillow needs an RGB surface in order to draw with alpha
        # blending, not RGBA
        self.surface = Image.new('RGB', (width, height), color=0)

        super(CanvasRenderer, self).resize(dims)

    ## def finalize(self):
    ##     cr = RenderContext(self, self.viewer, self.surface)
    ##     self.draw_vector(cr)

    def render_image(self, rgb_arr, order, win_coord):
        """Render the image represented by (data) at (win_coord)
        in the pixel space.
        *** internal method-- do not use ***
        """
        if self.surface is None:
            return
        self.logger.debug("redraw surface")

        # get window contents as a buffer and paste it into the PIL surface
        # TODO: allow greater bit depths when support is better in PIL
        p_image = Image.fromarray(rgb_arr)

        if self.surface is None or p_image.size != self.surface.size:
            # window size must have changed out from underneath us!
            width, height = self.viewer.get_window_size()
            self.resize((width, height))
            if p_image.size != self.surface.size:
                raise render.RenderError("Rendered image does not match window size")

        self.surface.paste(p_image)

    def get_surface_as_array(self, order=None):
        if self.surface is None:
            raise render.RenderError("No PIL surface defined")

        # TODO: could these have changed between the time that self.surface
        # was last updated?
        wd, ht = self.dims[:2]

        # Get PIL surface as a numpy array
        arr8 = np.frombuffer(self.surface.tobytes(), dtype=np.uint8)
        arr8 = arr8.reshape((ht, wd, 3))

        # adjust according to viewer's needed order
        arr8 = self.reorder(order, arr8, 'RGB')
        return arr8

    def setup_cr(self, shape):
        cr = RenderContext(self, self.viewer, self.surface)
        return cr

    def get_dimensions(self, shape):
        cr = self.setup_cr(shape)
        cr.set_font_from_shape(shape)
        return cr.text_extents(shape.text)

    def text_extents(self, text, font):
        cr = RenderContext(self, self.viewer, self.surface)
        cr.set_font(font.fontname, font.fontsize)
        return cr.text_extents(text)

#END



#
# cmap.py -- mode for setting color map
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""CMap Mode enables bindings that can adjust the color and intensity
maps of an image in a Ginga image viewer.

Enter the mode by
-----------------
* Space, then "y"

Exit the mode by
----------------
* Esc

Default bindings in mode
------------------------
* Y : reset colormap to default (gray)
* r : restore color map (undo any color map inversion, rotation or stretch)
* I : invert current color map (preserving any rotation or stretch)
* up arrow : set previous colormap in list (preserves any rotation or stretch)
* down arrow : set next colormap in list (preserves any rotation or stretch)
* c : overlay a visible colormap on the image
* i : restore the default intensity map to "ramp", the default
* j, left arrow : set previous intensity map in list
* k, right arrow : set next intensity map in list
* scroll : choose color map from list
* Ctrl + scroll : choose intensity map from list
* pan gesture: choose color map from list (hint: use up/down arrows keys
  to finalize selection)
* left drag : rotate current color map
* right click : restore color map (same as "r")

"""
import numpy as np

from ginga import cmap, imap
from ginga.modes.mode_base import Mode


class CMapMode(Mode):

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_cmap=['__y', None, None],

            kp_cmap_reset=['Y', 'cmap+Y'],
            kp_cmap_restore=['cmap+r'],
            kp_cmap_invert=['I', 'cmap+I'],
            kp_cmap_prev=['cmap+up', 'cmap+b'],
            kp_cmap_next=['cmap+down', 'cmap+n'],
            kp_toggle_cbar=['cmap+c'],
            kp_imap_reset=['cmap+i'],
            kp_imap_prev=['cmap+left', 'cmap+j'],
            kp_imap_next=['cmap+right', 'cmap+k'],

            sc_cmap=['cmap+scroll'],
            sc_imap=['cmap+ctrl+scroll'],

            ms_cmap_rotate=['cmap+left'],
            ms_cmap_restore=['cmap+right'],

            pa_cmap_cycle=['cmap+pan'])

    def __str__(self):
        return 'cmap'

    @property
    def cancmap(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('cmap')

    def start(self):
        pass

    def stop(self):
        self.onscreen_message(None)

    def restore_colormap(self, viewer, msg=True):
        msg = self.settings.get('msg_cmap', msg)
        rgbmap = viewer.get_rgbmap()
        rgbmap.restore_cmap()
        if msg:
            self.onscreen_message("Restored color map", delay=0.5)
        return True

    def _rotate_colormap(self, viewer, x, y, mode):
        win_wd, win_ht = viewer.get_window_size()

        # translate X cursor position as a percentage of the window
        # width into a shifting factor
        shift_pct = np.clip(x, 0, win_wd) / float(win_wd)
        self.logger.debug("rotating color map by %.2f pct" % (shift_pct))

        viewer.get_settings().set(color_map_rot_pct=shift_pct)

    def _cycle_cmap(self, viewer, msg, direction='down'):
        msg = self.settings.get('msg_cmap', msg)
        rgbmap = viewer.get_rgbmap()
        cm = rgbmap.get_cmap()
        cmapname = cm.name
        cmapnames = cmap.get_names()
        idx = cmapnames.index(cmapname)
        if direction == 'down':
            idx = (idx + 1) % len(cmapnames)
        else:
            idx = idx - 1
            if idx < 0:
                idx = len(cmapnames) - 1
        cmapname = cmapnames[idx]
        viewer.set_color_map(cmapname)
        if msg:
            self.onscreen_message("Color map: %s" % (cmapname),
                                  delay=1.0)

    def _reset_cmap(self, viewer, msg):
        msg = self.settings.get('msg_cmap', msg)
        # default
        with viewer.suppress_redraw:
            t_ = viewer.get_settings()
            cmapname = 'gray'
            t_.set(color_map=cmapname, intensity_map='ramp',
                   color_map_invert=False, color_map_rot_pct=0.0)
            if msg:
                self.onscreen_message("Color map: %s" % (cmapname),
                                      delay=1.0)

    def _invert_cmap(self, viewer, msg):
        msg = self.settings.get('msg_cmap', msg)
        viewer.invert_color_map()
        if msg:
            self.onscreen_message("Inverted color map", delay=1.0)

    def _cycle_imap(self, viewer, msg, direction='down'):
        msg = self.settings.get('msg_imap', msg)
        rgbmap = viewer.get_rgbmap()
        im = rgbmap.get_imap()
        imapname = im.name
        imapnames = imap.get_names()
        idx = imapnames.index(imapname)
        if direction == 'down':
            idx = (idx + 1) % len(imapnames)
        else:
            idx = idx - 1
            if idx < 0:
                idx = len(imapnames) - 1
        imapname = imapnames[idx]
        viewer.set_intensity_map(imapname)
        if msg:
            self.onscreen_message("Intensity map: %s" % (imapname),
                                  delay=1.0)

    def _reset_imap(self, viewer, msg):
        msg = self.settings.get('msg_imap', msg)
        # default
        imapname = 'ramp'
        viewer.set_intensity_map(imapname)
        if msg:
            self.onscreen_message("Intensity map: %s" % (imapname),
                                  delay=1.0)

    #####  KEYBOARD ACTION CALLBACKS #####

    def kp_cmap_reset(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._reset_cmap(viewer, msg)

    def kp_cmap_restore(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self.restore_colormap(viewer, msg)
        return True

    def kp_cmap_invert(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._invert_cmap(viewer, msg)
        return True

    def kp_cmap_prev(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._cycle_cmap(viewer, msg, direction='up')
        return True

    def kp_cmap_next(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._cycle_cmap(viewer, msg, direction='down')
        return True

    def kp_toggle_cbar(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        canvas = viewer.get_private_canvas()
        # canvas already has a color bar?
        objs = list(canvas.get_objects_by_kinds(('colorbar', 'drawablecolorbar')))
        tf = (len(objs) == 0)
        viewer.show_color_bar(tf)
        return True

    def kp_imap_reset(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._reset_imap(viewer, msg)
        return True

    def kp_imap_prev(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._cycle_imap(viewer, msg, direction='up')
        return True

    def kp_imap_next(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._cycle_imap(viewer, msg, direction='down')
        return True

    #####  SCROLL ACTION CALLBACKS #####

    def sc_cmap(self, viewer, event, msg=True):
        """Interactively change the color map by scrolling.
        """
        if not self.cancmap:
            return False
        event.accept()
        direction = self.get_direction(event.direction)
        self._cycle_cmap(viewer, msg, direction=direction)

    def sc_imap(self, viewer, event, msg=True):
        """Interactively change the intensity map by scrolling.
        """
        if not self.cancmap:
            return False
        event.accept()
        direction = self.get_direction(event.direction)
        self._cycle_imap(viewer, msg, direction=direction)

    #####  MOUSE ACTION CALLBACKS #####

    def ms_cmap_rotate(self, viewer, event, data_x, data_y, msg=True):
        """Shift the colormap by dragging the cursor left or right.
        Stretch the colormap by dragging the cursor up or down.
        """
        if not self.cancmap:
            return False
        event.accept()
        msg = self.settings.get('msg_cmap', msg)

        x, y = self.get_win_xy(viewer)

        if event.state == 'move':
            self._rotate_colormap(viewer, x, y, 'preview')

        elif event.state == 'down':
            self._start_x, self._start_y = x, y
            if msg:
                self.onscreen_message("Rotate colormap (drag mouse L/R)",
                                      delay=1.0)
        else:
            self.onscreen_message(None)

    def ms_cmap_restore(self, viewer, event, data_x, data_y, msg=True):
        """An interactive way to restore the colormap settings after
        a rotate or invert operation.
        """
        if not self.cancmap:
            return False
        event.accept()
        if event.state == 'down':
            self.restore_colormap(viewer, msg)

    ##### GESTURE ACTION CALLBACKS #####

    def pa_cmap_cycle(self, viewer, event, msg=True):
        """Change the color map by a pan gesture.
        (the back end must support gestures)
        """
        if not self.cancmap:
            return False
        event.accept()
        event = self._pa_synth_scroll_event(event)
        if event.state == 'move':
            rev = self.settings.get('zoom_scroll_reverse', False)
            direction = self.get_direction(event.direction, rev=rev)
            self._cycle_cmap(viewer, msg, direction=direction)

            return True
        return False



#
# PlotView.py -- base class for plot viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import logging
from io import BytesIO
import numpy as np

from ginga import Mixins
from ginga.misc import Callback, Settings, Bunch
from ginga.AstroImage import AstroImage
from ginga.plot.Plotable import Plotable
from ginga.canvas.CanvasObject import get_canvas_types
from ginga.cursors import cursor_info
from ginga import events, AutoCuts

try:
    from matplotlib.figure import Figure
    from matplotlib.backend_tools import Cursors

    from ginga.gw.Plot import PlotWidget
    have_mpl = True
except ImportError:
    have_mpl = False

__all__ = ['PlotViewBase', 'PlotViewEvent', 'PlotViewGw']


class PlotViewBase(Callback.Callbacks):
    """A Ginga viewer for displaying 2D plots using matplotlib.
    """

    vname = 'Ginga Plot'
    vtypes = [AstroImage, Plotable]

    @classmethod
    def viewable(cls, dataobj):
        """Test whether `dataobj` is viewable by this viewer."""
        if isinstance(dataobj, Plotable):
            return True
        if not isinstance(dataobj, AstroImage):
            return False
        shp = list(dataobj.shape)
        # comment this check to be able to view 2D images with this viewer
        if 0 in shp or len(shp) != 1:
            return False
        return True

    def __init__(self, logger=None, settings=None, figure=None):
        Callback.Callbacks.__init__(self)

        if not have_mpl:
            raise ImportError("Install 'matplotlib' to use this viewer")

        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.Logger('PlotView')
        self.needs_scrolledview = True

        # Create settings and set defaults
        # NOTE: typically inheriting channel settings
        if settings is None:
            settings = Settings.SettingGroup(logger=self.logger)
        self.settings = settings
        self.t_ = settings

        self.t_.add_defaults(plot_bg='white',
                             plot_save_suffix='.png',
                             plot_dpi=100, plot_save_dpi=100,
                             plot_title_fontsize=14,
                             plot_axis_fontsize=14,
                             plot_limits=None,
                             plot_range=None,
                             plot_enter_focus=True,
                             plot_zoom_rate=1.1)

        # pan position
        self.t_.add_defaults(plot_pan=(0.0, 0.0))
        self.t_.get_setting('plot_pan').add_callback('set', self.pan_change_cb)

        # for axis scaling
        self.t_.add_defaults(plot_dist_axis=('linear', 'linear'))
        for name in ['plot_dist_axis']:
            self.t_.get_setting(name).add_callback('set', self.axis_dist_change_cb)

        # plot markers
        self.t_.add_defaults(plot_show_marker=False, plot_marker_size=5,
                             plot_marker_width=0.35,
                             plot_marker_color='blue',
                             plot_marker_style='o')
        for name in ['plot_show_marker', 'plot_marker_size',
                     'plot_marker_width', 'plot_marker_color',
                     'plot_marker_style']:
            self.t_.get_setting(name).add_callback('set',
                                                   lambda setting, value: self.replot())

        # TODO
        width, height = 500, 500

        if figure is None:
            figure = Figure()
            dpi = figure.get_dpi()
            if dpi is None or dpi < 0.1:
                dpi = self.t_['plot_dpi']
            wd_in, ht_in = float(width) / dpi, float(height) / dpi
            figure.set_size_inches(wd_in, ht_in)
        self.figure = figure
        if hasattr(self.figure, 'set_tight_layout'):
            self.figure.set_tight_layout(True)

        self._dataobj = None
        self._data_limits = None
        self.rgb_order = 'RGBA'

        # for debugging
        self.name = str(self)
        # cursors
        self.cursor = {}

        self.plot_w = PlotWidget(self)

        self.artist_dct = dict()
        bg = self.settings.get('plot_bg', 'white')
        self.ax = self.figure.add_subplot(111, facecolor=bg)
        self.ax.grid(True)

        self.dc = get_canvas_types()
        klass = AutoCuts.get_autocuts('zscale')
        self.autocuts = klass(self.logger)

        # For callbacks
        for name in ['image-set', 'image-unset',
                     'limits-set', 'range-set', 'redraw',
                     'configure']:
            self.enable_callback(name)

    def get_figure(self):
        return self.figure

    def get_widget(self):
        # same as self.plot_w
        return self.figure.canvas

    def add_axis(self, **kwdargs):
        self.ax = self.figure.add_subplot(111, **kwdargs)
        return self.ax

    def get_axis(self):
        return self.ax

    def get_settings(self):
        return self.settings

    def get_logger(self):
        return self.logger

    def initialize_channel(self, fv, channel):
        # no housekeeping to do (for now) on our part, just override to
        # suppress the logger warning
        pass

    def set_dataobj(self, dataobj):
        """Set 1D data to be displayed.

        If there is no error, the ``'image-unset'`` and ``'image-set'``
        callbacks will be invoked.

        Parameters
        ----------
        dataobj : `~ginga.AstroImage.AstroImage` or `~ginga.plot.Plotable.Plotable`
            AstroImage object or Plotable.

        """
        if not self.viewable(dataobj):
            raise ValueError("Wrong type of object to load: %s" % (
                str(type(dataobj))))

        old_dataobj = self._dataobj
        if old_dataobj is not None:
            self.make_callback('image-unset', old_dataobj)
        self._dataobj = dataobj

        self.clear()

        if isinstance(dataobj, AstroImage):
            dataobj.add_callback('modified', lambda dataobj: self.replot())
            self.show_image(dataobj)

        elif isinstance(dataobj, Plotable):
            dataobj.add_callback('modified', lambda dataobj: self.replot())
            self.show_plotable(dataobj)

        self.zoom_fit()

        self.make_callback('image-set', dataobj)

    def get_dataobj(self):
        return self._dataobj

    def clear(self, redraw=True):
        """Clear plot display."""
        self.clear_data()
        self.t_['plot_range'] = None
        if redraw:
            self.redraw()

    def clear_data(self):
        self.logger.debug('clearing viewer...')
        self.artist_dct = dict()
        self.ax.set_aspect('auto')
        self.ax.cla()
        self.t_['plot_limits'] = None
        self._data_limits = None

    def remove_artist_category(self, artist_category, redraw=True):
        artists = self.artist_dct.setdefault(artist_category, [])
        for artist in artists:
            try:
                artist.remove()
            except Exception as e:
                pass
        if redraw:
            self.redraw()

    def set_titles(self, title=None, x_axis=None, y_axis=None, redraw=True):
        if x_axis is not None:
            self.ax.set_xlabel(x_axis)
        if y_axis is not None:
            self.ax.set_ylabel(y_axis)
        if title is not None:
            self.ax.set_title(title)
        ax = self.ax
        ax.title.set_fontsize(self.t_['plot_title_fontsize'])
        for item in ([ax.xaxis.label, ax.yaxis.label] +
                     ax.get_xticklabels() + ax.get_yticklabels()):
            item.set_fontsize(self.t_['plot_axis_fontsize'])

        # Make x axis labels a little more readable
        lbls = self.ax.xaxis.get_ticklabels()
        for lbl in lbls:
            lbl.set(rotation=45, horizontalalignment='right')

        if redraw:
            self.redraw()

    def set_grid(self, tf):
        self.ax.grid(tf)

    def _record_limits(self, x_data, y_data):
        x_min, x_max = np.nanmin(x_data), np.nanmax(x_data)
        y_min, y_max = np.nanmin(y_data), np.nanmax(y_data)

        adjusted = False
        if self._data_limits is None:
            x_lo, y_lo, x_hi, y_hi = x_min, y_min, x_max, y_max
            adjusted = True
        else:
            (x_lo, y_lo), (x_hi, y_hi) = self._data_limits
            if x_min < x_lo:
                x_lo, adjusted = x_min, True
            if x_max > x_hi:
                x_hi, adjusted = x_max, True
            if y_min < y_lo:
                y_lo, adjusted = y_min, True
            if y_max > y_hi:
                y_hi, adjusted = y_max, True

        if adjusted:
            self._data_limits = [(x_lo, y_lo), (x_hi, y_hi)]

    def _plot_line(self, x_data, y_data, artist_category='default', name=None,
                   linewidth=1, linestyle='-', color='black',
                   alpha=1.0):
        """Simple line plot."""

        plt_kw = {'lw': linewidth,
                  'ls': linestyle,
                  'color': color,
                  'alpha': alpha,
                  'antialiased': True,
                  }
        if self.t_['plot_show_marker']:
            plt_kw.update({'marker': self.t_['plot_marker_style'],
                           'ms': self.t_['plot_marker_size'],
                           'mew': self.t_['plot_marker_width'],
                           'mfc': self.t_['plot_marker_color'],
                           'mec': self.t_['plot_marker_color']})

        try:
            artists = self.artist_dct.setdefault(artist_category, [])
            line, = self.ax.plot(x_data, y_data, **plt_kw)
            artists.append(line)

            # adjust limits if necessary
            self._record_limits(x_data, y_data)

        except Exception as e:
            self.logger.error(str(e), exc_info=True)

    def _plot_text(self, x, y, text, artist_category='default', rot_deg=0,
                   linewidth=1, color='black', alpha=1.0,
                   font='sans', fontsize=12):
        artists = self.artist_dct.setdefault(artist_category, [])
        fontdict = dict(color=color, family=font, size=fontsize)
        text = self.ax.text(x, y, text, color=color, rotation=rot_deg,
                            fontdict=fontdict, clip_on=True)
        artists.append(text)

        # adjust limits if necessary
        self._record_limits(x, y)

    def _plot_normimage(self, dataobj, artist_category='default', rot_deg=0):
        artists = self.artist_dct.setdefault(artist_category, [])
        # Get the data extents
        x0, y0 = 0, 0
        data_np = dataobj.get_data()
        y1, x1 = data_np.shape[:2]
        # flipx, flipy, swapxy = self.get_transforms()
        # if swapxy:
        #     x0, x1, y0, y1 = y0, y1, x0, x1

        locut, hicut = self.autocuts.calc_cut_levels_data(data_np)

        extent = (x0, x1, y0, y1)
        img = self.ax.imshow(data_np, origin='lower',
                             interpolation='none',
                             norm='linear',  # also 'log',
                             vmin=locut, vmax=hicut,
                             cmap='gray',
                             extent=extent)
        artists.append(img)

        # adjust limits if necessary
        self._record_limits(np.array((x0, x1)), np.array((y0, y1)))

    def set_limits(self, limits):
        self.settings.set(plot_limits=limits)

        # NOTE: make compatible callback to image viewer
        self.make_callback('limits-set', limits)

    def get_limits(self):
        limits = self.settings['plot_limits']

        if limits is None:
            # No user defined limits. Return limits set from data points.
            if self._data_limits is None:
                return [[0.0, 0.0], [0.0, 0.0]]

            limits = self._data_limits.copy()

        return limits

    def reset_limits(self):
        """Reset the bounding box of the viewer extents.

        Parameters
        ----------
        None
        """
        self.t_.set(plot_limits=None)

    def get_pan_bbox(self):
        """Get the coordinates in the actual data corresponding to the
        area shown in the display for the current zoom level and pan.

        Returns
        -------
        points : list
            Coordinates in the form of
            ``[(x0, y0), (x1, y1), (x2, y2), (x3, y3)]``
            from lower-left to lower-right.

        """
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()
        return [(x_lo, y_lo), (x_hi, y_lo), (x_hi, y_hi), (x_lo, y_hi)]

    def set_ranges(self, x_range=None, y_range=None):
        adjusted = False
        ranges = self.get_ranges()
        (x_lo, x_hi), (y_lo, y_hi) = ranges

        if x_range is not None:
            x_lo, x_hi = x_range
            ranges[0] = [x_lo, x_hi]
            adjusted = True
            self.ax.set_xlim(x_lo, x_hi)

        if y_range is not None:
            y_lo, y_hi = y_range
            ranges[1] = [y_lo, y_hi]
            adjusted = True
            self.ax.set_ylim(y_lo, y_hi)

        if adjusted:
            self.t_['plot_range'] = ranges
            self.make_callback('range-set', ranges.copy())

            _pan_x, _pan_y = self.t_['plot_pan'][:2]
            pan_x, pan_y = (x_lo + x_hi) * 0.5, (y_lo + y_hi) * 0.5
            if not np.isclose(_pan_x, pan_x) or not np.isclose(_pan_y, pan_y):
                # need to set the pan position because range is changing
                # not according to pan position
                #self.t_['plot_pan'] = (pan_x, pan_y)
                self.set_pan(pan_x, pan_y)

            #self.make_callback('range-set', ranges.copy())

            self.redraw()

    def get_ranges(self):
        ranges = self.settings['plot_range']
        if ranges is None:
            (x_lo, y_lo), (x_hi, y_hi) = self.get_limits()
            return [[x_lo, x_hi], [y_lo, y_hi]]

        #return [(x_lo, x_hi), (y_lo, y_hi)]
        return ranges.copy()

    def zoom_fit(self, axis='xy', no_reset=False):
        """Zoom to fit display window.
        Pan the image and scale the view to fit the size of the set
        limits (usually set to the image size).  Parameter `axis` can
        be used to set which axes are allowed to be scaled;
        Also see :meth:`zoom_to`.

        Parameters
        ----------
        axis : str
            One of: 'x', 'y', or 'xy' (default).

        no_reset : bool
            Do not reset ``autozoom`` setting.

        """
        # calculate actual width of the limits/image, considering rotation
        (x_min, y_min), (x_max, y_max) = self.get_limits()
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()

        # account for t_[scale_x/y_base]
        if axis in ['x', 'xy']:
            x_lo, x_hi = x_min, x_max
        if axis in ['y', 'xy']:
            y_lo, y_hi = y_min, y_max

        self.set_ranges(x_range=(x_lo, x_hi), y_range=(y_lo, y_hi))

        if self.t_['autozoom'] == 'once':
            self.t_.set(autozoom='off')

    def set_dist_axis(self, x_axis=None, y_axis=None):
        if x_axis is None:
            x_axis = self.t_['plot_dist_axis'][0]
        if y_axis is None:
            y_axis = self.t_['plot_dist_axis'][1]

        self.t_.set(plot_dist_axis=(x_axis, y_axis))

    def axis_dist_change_cb(self, setting, value):
        x_axis, y_axis = value
        if x_axis is not None:
            self.ax.set_xscale(x_axis)
        if y_axis is not None:
            self.ax.set_yscale(y_axis)

        self.redraw()

    def redraw_now(self, whence=0):
        self.figure.canvas.draw()

        self.make_callback('redraw', whence)

    def redraw(self, whence=0):
        self.redraw_now(whence=whence)

    def render_canvas(self, canvas):
        for obj in canvas.objects:
            if isinstance(obj, self.dc.Path):
                data_x, data_y = obj.points.T
                self._plot_line(data_x, data_y, linewidth=obj.linewidth,
                                linestyle=obj.linestyle, color=obj.color,
                                alpha=obj.alpha)

            elif isinstance(obj, self.dc.Line):
                data_x, data_y = [obj.x1, obj.x2], [obj.y1, obj.y2]
                self._plot_line(data_x, data_y, linewidth=obj.linewidth,
                                linestyle=obj.linestyle, color=obj.color,
                                alpha=obj.alpha)

            elif isinstance(obj, self.dc.Text):
                self._plot_text(obj.x, obj.y, text=obj.text,
                                rot_deg=obj.rot_deg,
                                color=obj.color, fontsize=obj.fontsize)

            elif isinstance(obj, self.dc.NormImage):
                image = obj.get_image()
                self._plot_normimage(image, rot_deg=obj.rot_deg)

            elif isinstance(obj, self.dc.Canvas):
                self.render_canvas(obj)

    def show_plotable(self, plotable):
        canvas = plotable.get_canvas()
        self.render_canvas(canvas)

        titles = plotable.titles
        self.set_titles(title=titles.title,
                        x_axis=titles.x_axis, y_axis=titles.y_axis,
                        redraw=False)
        self.set_grid(plotable.grid)

        self.redraw()

    def show_image(self, image):
        data_np = image.get_data()

        self.set_grid(False)
        if len(data_np.shape) == 1:
            x_data = np.arange(len(data_np))
            self.set_titles(title=image.get('name', "NoName 1D data"),
                            x_axis="Index", y_axis="Value",
                            redraw=False)
            self._plot_line(x_data, data_np)

        elif len(data_np.shape) == 2:
            self.set_titles(title=image.get('name', "NoName 2D data"),
                            x_axis="X", y_axis="Y",
                            redraw=False)
            self._plot_normimage(image)

        self.redraw()

    def replot(self):
        dataobj = self._dataobj
        if dataobj is None:
            return

        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()
        self.clear_data()

        if isinstance(dataobj, AstroImage):
            self.show_image(dataobj)

        elif isinstance(dataobj, Plotable):
            self.show_plotable(dataobj)

        self.set_ranges(x_range=(x_lo, x_hi), y_range=(y_lo, y_hi))

    def get_pan(self, coord='data'):
        """Get pan positions.

        Parameters
        ----------
        coord : {'data'}
            For compatibility with the image viewer.

        Returns
        -------
        positions : tuple
            X and Y positions, in that order.

        """
        if coord != 'data':
            raise ValueError("`coord` must be 'data' with this viewer")
        return self.t_['plot_pan']

    def set_pan(self, pan_x, pan_y, coord='data', no_reset=False):
        """Set pan position.

        Parameters
        ----------
        pan_x, pan_y : float
            Pan positions in X and Y.

        coord : {'data', 'wcs'}
            Indicates whether the given pan positions are in data or WCS space.

        no_reset : bool
            Do not reset ``autocenter`` setting.

        """
        self.t_.set(plot_pan=(pan_x, pan_y))

    def pan_change_cb(self, setting, value):
        pan_x, pan_y = value

        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()
        x_delta, y_delta = (x_hi - x_lo) * 0.5, (y_hi - y_lo) * 0.5
        x1, x2 = pan_x - x_delta, pan_x + x_delta
        y1, y2 = pan_y - y_delta, pan_y + y_delta

        self.set_ranges(x_range=(x1, x2), y_range=(y1, y2))

    def panset_xy(self, data_x, data_y, no_reset=False):
        """Similar to :meth:`set_pan`, except that input pan positions
        are always in data space.

        """
        self.set_pan(data_x, data_y, coord='data', no_reset=no_reset)

    def pan_delta_px(self, x_delta_px, y_delta_px):
        """Pan by a delta in X and Y specified in pixels.

        Parameters
        ----------
        x_delta_px : float
            Delta pixels in X

        y_delta_px : float
            Delta pixels in Y

        """
        pan_x, pan_y = self.get_pan(coord='data')[:2]
        pan_x += x_delta_px
        pan_y += y_delta_px
        self.panset_xy(pan_x, pan_y)

    def panset_pct(self, pct_x, pct_y):
        """Similar to :meth:`set_pan`, except that pan positions
        are determined by multiplying data dimensions with the given
        scale factors, where 1 is 100%.

        """
        xy_mn, xy_mx = self.get_limits()
        pan_x, pan_y = self.get_pan()[:2]

        if pct_x is not None:
            pan_x = (xy_mn[0] + xy_mx[0]) * pct_x
        if pct_y is not None:
            pan_y = (xy_mn[1] + xy_mx[1]) * pct_y

        self.panset_xy(pan_x, pan_y)

    def calc_pan_pct(self, pad=0, min_pct=0.0, max_pct=0.9):
        """Calculate values for vertical/horizontal panning by percentages
        from the current pan position.  Mostly used by scrollbar callbacks.

        Parameters
        ----------
        pad : int (optional, defaults to 0)
            a padding amount in pixels to add to the limits when calculating

        min_pct : float (optional, range 0.0:1.0, defaults to 0.0)
        max_pct : float (optional, range 0.0:1.0, defaults to 0.9)

        Returns
        -------
        res : `~ginga.misc.Bunch.Bunch`
            calculation results, which include the following attributes:
            - rng_x : the range of X of the limits (including padding)
            - rng_y : the range of Y of the limits (including padding)
            - vis_x : the visually shown range of X in the viewer
            - vis_y : the visually shown range of Y in the viewer
            - thm_pct_x : the length of a X scrollbar arm as a ratio
            - thm_pct_y : the length of a Y scrollbar arm as a ratio
            - pan_pct_x : the pan position of X as a ratio
            - pan_pct_y : the pan position of Y as a ratio

        """
        limits = self.get_limits()
        pan_x, pan_y = self.get_pan()[:2]

        # calculate the corners of the entire image in unscaled cartesian
        mxwd, mxht = limits[1][:2]
        mxwd, mxht = mxwd + pad, mxht + pad
        mnwd, mnht = limits[0][:2]
        mnwd, mnht = mnwd - pad, mnht - pad

        arr = np.array([(mnwd, mnht), (mxwd, mnht),
                        (mxwd, mxht), (mnwd, mxht)],
                       dtype=float)
        x, y = arr.T
        x, y = x - pan_x, y - pan_y
        x_min, x_max = np.min(x), np.max(x)
        y_min, y_max = np.min(y), np.max(y)

        # this is the range of X and Y of the entire image
        # in the viewer (unscaled)
        rng_x, rng_y = abs(x_max - x_min), abs(y_max - y_min)

        # this is the *visually shown* range of X and Y
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()
        vis_x, vis_y = x_hi - x_lo, y_hi - y_lo
        arr = np.array([(x_lo, y_lo), (x_hi, y_hi)])
        x, y = arr.T
        x, y = x - pan_x, y - pan_y

        # calculate the length of the slider arms as a ratio
        xthm_pct = max(min_pct, min(vis_x / rng_x, max_pct))
        ythm_pct = max(min_pct, min(vis_y / rng_y, max_pct))

        # calculate the pan position as a ratio
        pct_x = min(max(0.0, abs(x_min) / rng_x), 1.0)
        pct_y = min(max(0.0, abs(y_min) / rng_y), 1.0)

        bnch = Bunch.Bunch(rng_x=rng_x, rng_y=rng_y, vis_x=vis_x, vis_y=vis_y,
                           thm_pct_x=xthm_pct, thm_pct_y=ythm_pct,
                           pan_pct_x=pct_x, pan_pct_y=pct_y)
        return bnch

    def pan_by_pct(self, pct_x, pct_y, pad=0):
        """Pan by a percentage of the data space. This method is designed
        to be called by scrollbar callbacks.

        Parameters
        ----------
        pct_x : float (range 0.0 : 1.0)
            Percentage in the X range to pan

        pct_y : float (range 0.0 : 1.0)
            Percentage in the Y range to pan

        pad : int (optional, defaults to 0)
            a padding amount in pixels to add to the limits when calculating

        """
        # Sanity check on inputs
        pct_x = np.clip(pct_x, 0.0, 1.0)
        pct_y = np.clip(pct_y, 0.0, 1.0)

        limits = self.get_limits()

        mxwd, mxht = limits[1][:2]
        mxwd, mxht = mxwd + pad, mxht + pad
        mnwd, mnht = limits[0][:2]
        mnwd, mnht = mnwd - pad, mnht - pad

        arr = np.array([(mnwd, mnht), (mxwd, mnht),
                        (mxwd, mxht), (mnwd, mxht)],
                       dtype=float)
        x, y = arr.T
        x_min, x_max = np.min(x), np.max(x)
        y_min, y_max = np.min(y), np.max(y)

        crd_x = x_min + (pct_x * (x_max - x_min))
        crd_y = y_min + (pct_y * (y_max - y_min))

        pan_x, pan_y = crd_x, crd_y
        self.logger.debug("crd=%f,%f pan=%f,%f" % (
            crd_x, crd_y, pan_x, pan_y))

        self.panset_xy(pan_x, pan_y)

    def zoom_plot(self, pct_x, pct_y, redraw=True):
        """Zoom the plot, keeping the pan position."""
        set_range = False
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()
        pan_x, pan_y = self.get_pan()[:2]

        if pct_x is not None:
            xrng = x_hi - x_lo
            xinc = (xrng * 0.5) * pct_x
            x_lo, x_hi = pan_x - xinc, pan_x + xinc
            set_range = True

        if pct_y is not None:
            yrng = y_hi - y_lo
            yinc = (yrng * 0.5) * pct_y
            y_lo, y_hi = pan_y - yinc, pan_y + yinc
            set_range = True

        if set_range:
            self.set_ranges(x_range=(x_lo, x_hi), y_range=(y_lo, y_hi))

    def zoom_plot_at_cursor(self, cur_x, cur_y, pct_x, pct_y):
        """Zoom the plot, keeping the position under the cursor."""
        set_range = False
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()

        if pct_x is not None:
            x_lo, x_hi = (cur_x - (cur_x - x_lo) * pct_x,
                          cur_x + (x_hi - cur_x) * pct_x)
            set_range = True

        if pct_y is not None:
            y_lo, y_hi = (cur_y - (cur_y - y_lo) * pct_y,
                          cur_y + (y_hi - cur_y) * pct_y)
            set_range = True

        if set_range:
            self.set_ranges(x_range=(x_lo, x_hi), y_range=(y_lo, y_hi))

    def get_axes_size_in_px(self):
        bbox = self.ax.get_window_extent().transformed(self.figure.dpi_scale_trans.inverted())
        width, height = bbox.width, bbox.height
        width *= self.figure.dpi
        height *= self.figure.dpi
        return (width, height)

    def get_rgb_image_as_buffer(self, output=None, format='png',
                                quality=90):
        """Get the current image shown in the viewer, with any overlaid
        graphics, in a file IO-like object encoded as a bitmap graphics
        file.

        This can be overridden by subclasses.

        Parameters
        ----------
        output : a file IO-like object or None
            open python IO descriptor or None to have one created

        format : str
            A string defining the format to save the image.  Typically
            at least 'jpeg' and 'png' are supported. (default: 'png')

        quality: int
            The quality metric for saving lossy compressed formats.

        Returns
        -------
        buffer : file IO-like object
            This will be the one passed in, unless `output` is None
            in which case a BytesIO obejct is returned

        """
        obuf = output
        if obuf is None:
            obuf = BytesIO()

        fig_dpi = self.settings.get('plot_save_dpi', 100)
        fig = self.get_figure()
        fig.savefig(obuf, dpi=fig_dpi)

        return obuf

    def save_rgb_image_as_file(self, filepath, format='png', quality=90):
        """Save the current image shown in the viewer, with any overlaid
        graphics, in a file with the specified format and quality.
        This can be overridden by subclasses.

        Parameters
        ----------
        filepath : str
            path of the file to write

        format : str
            See :meth:`get_rgb_image_as_buffer`.

        quality: int
            See :meth:`get_rgb_image_as_buffer`.

        """
        plot_ext = self.settings.get('plot_save_suffix', f'.{format}')
        if not filepath.endswith(plot_ext):
            filepath = filepath + plot_ext

        fig_dpi = self.settings.get('plot_save_dpi', 100)
        fig = self.get_figure()
        fig.savefig(filepath, dpi=fig_dpi)

    def get_cursor(self, cname):
        """Get the cursor stored under the name.
        This can be overridden by subclasses, if necessary.

        Parameters
        ----------
        cname : str
            name of the cursor to return.

        """
        return self.cursor[cname]

    def define_cursor(self, cname, cursor):
        """Define a viewer cursor under a name.  Does not change the
        current cursor.

        Parameters
        ----------
        cname : str
            name of the cursor to define.

        cursor : object
            a cursor object in the back end's toolkit

        `cursor` is usually constructed from `make_cursor`.
        """
        self.cursor[cname] = cursor

    def switch_cursor(self, cname):
        """Switch the viewer's cursor to the one defined under a name.

        Parameters
        ----------
        cname : str
            name of the cursor to switch to.

        """
        cursor = self.get_cursor(cname)
        self.figure.canvas.set_cursor(cursor)

    def __str__(self):
        return "PlotViewBase"


class PlotViewEvent(Mixins.UIMixin, PlotViewBase):

    def __init__(self, logger=None, settings=None, figure=None):
        PlotViewBase.__init__(self, logger=logger, settings=settings,
                              figure=figure)
        Mixins.UIMixin.__init__(self)

        # for interactive features
        self.can = Bunch.Bunch(zoom=False, pan=False)

        # for qt key handling
        self._keytbl = {
            'shift': 'shift_l',
            'control': 'control_l',
            'alt': 'alt_l',
            'win': 'super_l',
            'meta': 'meta_right',
            '`': 'backquote',
            '"': 'doublequote',
            "'": 'singlequote',
            '\\': 'backslash',
            ' ': 'space',
            'enter': 'return',
            'pageup': 'page_up',
            'pagedown': 'page_down',
        }
        # For callbacks
        #for name in []:
        #    self.enable_callback(name)

        self.last_data_x, self.last_data_y = 0, 0
        self.connect_ui()

        # enable interactivity in the plot
        self.__enable(zoom=True, pan=True)

    def __enable(self, pan=False, zoom=False):
        # NOTE: don't use this interface!  Likely to change!!!
        """If `pan` is True, enable interactive panning in the plot by a
        middle click.  If `zoom` is True , enable interactive zooming in
        the plot by scrolling.
        """
        self.can.update(dict(pan=pan, zoom=zoom))
        self.ui_set_active(True, viewer=self)
        if pan or zoom:
            if pan:
                self.add_callback('button-press', self.plot_do_pan)
            if zoom:
                self.add_callback('scroll', self.plot_do_zoom)

    def plot_do_zoom(self, cb_obj, event):
        """Can be set as the callback function for the 'scroll'
        event to zoom the plot.
        """
        if not self.can.zoom:
            return

        # Matplotlib only gives us the number of steps of the scroll,
        # positive for up and negative for down.
        if event.amount > 0:
            delta = self.t_['plot_zoom_rate'] ** -2
        elif event.amount < 0:
            delta = self.t_['plot_zoom_rate'] ** 2

        delta_x = delta_y = delta
        if 'ctrl' in event.modifiers:
            # only horizontal
            delta_y = 1.0
        elif 'shift' in event.modifiers or 'alt' in event.modifiers:
            # only vertical
            # (shift works on Linux, but not Mac; alt works on Mac but
            #  not Linux....Grrr)
            delta_x = 1.0

        if 'meta' in event.modifiers or 'cmd' in event.modifiers:
            # cursor position
            cur_x, cur_y = event.data_x, event.data_y
            if None not in [cur_x, cur_y]:
                self.zoom_plot_at_cursor(cur_x, cur_y, delta_x, delta_y)
        else:
            self.zoom_plot(delta_x, delta_y)

        return True

    def plot_do_pan(self, cb_obj, event):
        """Can be set as the callback function for the 'button-press'
        event to pan the plot with middle-click.
        """
        cur_x, cur_y = event.data_x, event.data_y

        if event.button == 0x2 or (event.button == 0x1 and
                                   'shift' in event.modifiers):
            if not self.can.pan:
                return
            if None not in [cur_x, cur_y]:
                self.set_pan(cur_x, cur_y)

        return True

    def connect_ui(self):
        canvas = self.figure.canvas
        if canvas is None:
            raise ValueError("matplotlib canvas is not yet created")
        connect = canvas.mpl_connect
        connect("motion_notify_event", self._plot_motion_notify)
        connect("button_press_event", self._plot_button_press)
        connect("button_release_event", self._plot_button_release)
        connect("scroll_event", self._plot_scroll)
        canvas.capture_scroll = True
        connect("figure_enter_event", self._plot_enter_cursor)
        connect("figure_leave_event", self._plot_leave_cursor)
        connect("key_press_event", self._plot_key_press)
        connect("key_release_event", self._plot_key_release)
        connect("resize_event", self._plot_resize)

        # Define cursors
        cursor_names = cursor_info.get_cursor_names()
        # TODO: handle other cursor types
        cross = Cursors.POINTER
        for curname in cursor_names:
            curinfo = cursor_info.get_cursor_info(curname)
            self.define_cursor(curinfo.name, cross)
        self.switch_cursor('pick')

    def __get_modifiers(self, event):
        return event.modifiers

    def __get_button(self, event):
        try:
            btn = 0x1 << (event.button.value - 1)
        except Exception:
            btn = 0
        return btn

    def transkey(self, keyname):
        self.logger.debug("matplotlib keyname='%s'" % (keyname))
        if keyname is None:
            return keyname
        key = keyname
        if 'shift+' in key:
            key = key.replace('shift+', '')
        if 'ctrl+' in key:
            key = key.replace('ctrl+', '')
        if 'alt+' in key:
            key = key.replace('alt+', '')
        if 'meta+' in key:
            key = key.replace('meta+', '')
        return self._keytbl.get(key, key)

    def get_key_table(self):
        return self._keytbl

    def __get_key(self, event):
        keyval = self.transkey(event.key)
        self.logger.debug("key event, mpl={}, key={}".format(event.key,
                                                             keyval))
        return keyval

    def _plot_scroll(self, event):
        button = self.__get_button(event)
        if event.button == 'up':
            direction = 0.0
        elif event.button == 'down':
            direction = 180.0
        amount = event.step
        modifiers = self.__get_modifiers(event)
        evt = events.ScrollEvent(viewer=self, button=button, state='scroll',
                                 mode=None, modifiers=modifiers,
                                 direction=direction, amount=amount,
                                 data_x=event.xdata, data_y=event.ydata)
        self.make_ui_callback('scroll', evt)

    def _plot_button_press(self, event):
        button = self.__get_button(event)
        modifiers = self.__get_modifiers(event)
        self.last_data_x, self.last_data_y = event.xdata, event.ydata
        evt = events.PointEvent(viewer=self, button=button, state='down',
                                mode=None, modifiers=modifiers,
                                data_x=event.xdata, data_y=event.ydata)
        self.make_ui_callback('button-press', evt)

    def _plot_button_release(self, event):
        button = self.__get_button(event)
        modifiers = self.__get_modifiers(event)
        self.last_data_x, self.last_data_y = event.xdata, event.ydata
        evt = events.PointEvent(viewer=self, button=button, state='up',
                                mode=None, modifiers=modifiers,
                                data_x=event.xdata, data_y=event.ydata)
        self.make_ui_callback('button-release', evt)

    def _plot_motion_notify(self, event):
        button = self.__get_button(event)
        modifiers = self.__get_modifiers(event)
        self.last_data_x, self.last_data_y = event.xdata, event.ydata
        evt = events.PointEvent(viewer=self, button=button, state='move',
                                mode=None, modifiers=modifiers,
                                data_x=event.xdata, data_y=event.ydata)
        self.make_ui_callback('motion', evt)

    def _plot_key_press(self, event):
        key = self.__get_key(event)
        modifiers = self.__get_modifiers(event)
        self.last_data_x, self.last_data_y = event.xdata, event.ydata
        evt = events.KeyEvent(viewer=self, key=key, state='down',
                              mode=None, modifiers=modifiers,
                              data_x=event.xdata, data_y=event.ydata)
        self.make_ui_callback('key-press', evt)

    def _plot_key_release(self, event):
        key = self.__get_key(event)
        modifiers = self.__get_modifiers(event)
        self.last_data_x, self.last_data_y = event.xdata, event.ydata
        evt = events.KeyEvent(viewer=self, key=key, state='up',
                              mode=None, modifiers=modifiers,
                              data_x=event.xdata, data_y=event.ydata)
        self.make_ui_callback('key-release', evt)

    def _plot_resize(self, event):
        wd, ht = event.width, event.height
        self.make_callback('configure', wd, ht)

    def _plot_enter_cursor(self, event):
        if self.t_['plot_enter_focus']:
            w = self.get_widget()
            if hasattr(w, 'setFocus'):
                # NOTE: this is a Qt call, not cross-backend
                # TODO: see if matplotlib has a backend independent way
                # to do this
                w.setFocus()

        self.make_ui_callback('enter')

    def _plot_leave_cursor(self, event):
        self.make_ui_callback('leave')

    def __str__(self):
        return "PlotViewEvent"


PlotViewGw = PlotViewEvent



#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import gettext
import os

localedir = os.path.abspath(os.path.dirname(__file__))
translate = gettext.translation('ginga', localedir, fallback=True)

# for testing
#_ = lambda x: x

_ = translate.gettext
_tr = translate.gettext



#
# plots.py -- classes for plots added to Ginga canvases.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import numpy as np

from ginga.canvas.CanvasObject import (CanvasObjectBase, _color,
                                       register_canvas_types,
                                       colors_plus_none)
from ginga.misc import Bunch
from ginga.canvas.types.layer import CompoundObject

from .basic import Path
from ginga.misc.ParamSet import Param


class XYPlot(CanvasObjectBase):
    """
    Plotable object that defines a single path representing an X/Y line plot.

    Like a Path, but has some optimization to reduce the actual numbers of
    points in the path, depending on the scale and pan of the viewer.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='linewidth', type=int, default=2,
                  min=0, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default: solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='black',
                  description="Color of text"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
        ]

    def __init__(self, name=None, color='black',
                 linewidth=1, linestyle='solid',
                 alpha=1.0, x_acc=None, y_acc=None, **kwargs):
        super(XYPlot, self).__init__(color=color, linewidth=linewidth,
                                     linestyle=linestyle, alpha=alpha,
                                     **kwargs)
        self.name = name
        self.kind = 'xyplot'
        self.x_func = None
        nul_arr = np.array([])
        if x_acc is not None:
            self.x_func = lambda arr: nul_arr if arr.size == 0 else x_acc(arr)
        if y_acc is None:
            y_acc = np.mean
        self.y_func = lambda arr: nul_arr if arr.size == 0 else y_acc(arr)

        self.points = np.copy(nul_arr)
        self.limits = np.array([(0.0, 0.0), (0.0, 0.0)])
        self.plot_xlim = (None, None)

        self.path = Path([], color=color, linewidth=linewidth,
                         linestyle=linestyle, alpha=alpha, coord='data')
        self.path.get_cpoints = self.get_cpoints

    def plot_xy(self, xpts, ypts):
        """Convenience function for plotting X and Y points that are in
        separate arrays.
        """
        self.plot(np.asarray((xpts, ypts)).T)

    def plot(self, points, limits=None):
        """Plot `points`, a list, tuple or array of (x, y) points.

        Parameter
        ---------
        points : array-like
            list, tuple or array of (x, y) points

        limits : array-like, optional
            array of (xmin, ymin), (xmax, ymax)

        Limits will be calculated if not passed in.
        """
        self.points = np.asarray(points)
        self.plot_xlim = (None, None)

        # set or calculate limits
        if limits is not None:
            # passing limits saves costly min/max calculation
            self.limits = np.asarray(limits)
        else:
            self._calc_limits(self.points)

    def _calc_limits(self, points):
        """Internal routine to calculate the limits of `points`.
        """
        # TODO: what should limits be if there are no points?
        if len(points) == 0:
            self.limits = np.array([[0.0, 0.0], [0.0, 0.0]])
        else:
            x_vals, y_vals = points.T
            self.limits = np.array([(x_vals.min(), y_vals.min()),
                                    (x_vals.max(), y_vals.max())])

    def calc_points(self, viewer, start_x, stop_x):
        """Called when recalculating our path's points.
        """
        # in case X axis is flipped
        start_x, stop_x = min(start_x, stop_x), max(start_x, stop_x)

        new_xlim = (start_x, stop_x)
        if new_xlim == self.plot_xlim:
            # X limits are the same, no need to recalculate points
            return

        self.plot_xlim = new_xlim

        points = self.get_data_points(points=self.points)
        if len(points) == 0:
            self.path.points = points
            return

        x_data, y_data = points.T

        # if we can determine the visible region shown on the plot
        # limit the points to those within the region
        if np.all(np.isfinite([start_x, stop_x])):
            idx = np.logical_and(x_data >= start_x, x_data <= stop_x)
            points = points[idx]

        if self.x_func is not None:
            # now find all points position in canvas X coord
            cpoints = self.get_cpoints(viewer, points=points)
            cx, cy = cpoints.T

            # Reduce each group of Y points that map to a unique X via a
            # function that reduces to a single value.  The desirable function
            # will depend on the function of the plot, but mean() would be a
            # sensible default
            _, i = np.unique(cx, return_index=True)
            gr_pts = np.split(points, i)
            x_data = np.array([self.x_func(a.T[0]) for a in gr_pts
                               if len(a) > 0])
            y_data = np.array([self.y_func(a.T[1]) for a in gr_pts
                               if len(a) > 0])
            assert len(x_data) == len(y_data)

            points = np.array((x_data, y_data)).T

        self.path.points = points

    def recalc(self, viewer):
        """Called when recalculating our path's points.
        """
        # select only points within range of the current pan/zoom
        bbox = viewer.get_pan_bbox()
        if bbox is None:
            self.path.points = []
            return

        start_x, stop_x = bbox[0][0], bbox[2][0]

        self.calc_points(viewer, start_x, stop_x)

    def get_cpoints(self, viewer, points=None, no_rotate=False):
        """Mostly internal routine used to calculate the native positions
        to draw the plot.
        """
        # If points are passed, they are assumed to be in data space
        if points is None:
            points = self.path.get_points()

        return viewer.tform['data_to_plot'].to_(points)

    def update_resize(self, viewer, dims):
        """Called when the viewer is resized."""
        self.recalc(viewer)

    def get_latest(self):
        """Get the latest (last) point on the plot.  Returns None if there
        are no points.
        """
        if len(self.points) == 0:
            return None
        return self.points[-1]

    def get_limits(self, lim_type):
        """Get the limits of the data or the visible part of the plot.

        If `lim_type` == 'data' returns the limits of all the data points.
        Otherwise returns the limits of the visibly plotted area.  Limits
        are returned in the form ((xmin, ymin), (xmax, ymax)), as an array.
        """
        if lim_type == 'data':
            # data limits
            return np.asarray(self.limits)

        # plot limits
        self.path.crdmap = self.crdmap
        if len(self.path.points) > 0:
            llur = self.path.get_llur()
            llur = [llur[0:2], llur[2:4]]
        else:
            llur = [(0.0, 0.0), (0.0, 0.0)]
        return np.asarray(llur)

    def draw(self, viewer):
        """Draw the plot.  Normally not called by the user, but by the viewer
        as needed.
        """
        self.path.crdmap = self.crdmap

        self.recalc(viewer)

        if len(self.path.points) > 0:
            self.path.draw(viewer)


class Axis(CompoundObject):
    """
    Base class for axis plotables.
    """
    def __init__(self, title=None, num_labels=4, font='sans',
                 fontsize=10.0):
        super(Axis, self).__init__()

        self.aide = None
        self.num_labels = num_labels
        self.title = title
        self.font = font
        self.fontsize = fontsize
        self.grid_alpha = 1.0
        self.format_value = self._format_value

    def register_decor(self, aide):
        self.aide = aide

    def _format_value(self, v):
        """Default formatter for XAxis labels.
        """
        return "%.4g" % v

    def set_grid_alpha(self, alpha):
        """Set the transparency (alpha) of the XAxis grid lines.
        `alpha` should be between 0.0 and 1.0
        """
        for i in range(self.num_labels):
            grid = self.grid[i]
            grid.alpha = alpha

    def get_data_xy(self, viewer, pt):
        arr_pts = np.asarray(pt)
        x, y = viewer.tform['data_to_plot'].from_(arr_pts).T[:2]
        flips = viewer.get_transforms()
        if flips[2]:
            x, y = y, x
        return (x, y)

    def get_title(self):
        titles_d = self.aide.get_axes_titles()
        return titles_d[self.kind]

    def add_plot(self, viewer, plot_src):
        # Axis objects typically do not need to do anything when a
        # plot is added--they recalculate labels in update_elements()
        pass

    def delete_plot(self, viewer, plot_src):
        # Axis objects typically do not need to do anything when a
        # plot is deleted--they recalculate labels in update_elements()
        pass


class XAxis(Axis):
    """
    Plotable object that defines X axis labels and grid lines.
    """
    def __init__(self, title=None, num_labels=4, font='sans',
                 fontsize=10.0):
        super(XAxis, self).__init__(title=title, num_labels=num_labels,
                                    font=font, fontsize=fontsize)

        self.kind = 'axis_x'
        self.txt_ht = 0
        self.title_wd = 0
        self.pad_px = 5

    def register_decor(self, aide):
        self.aide = aide

        # add X grid
        self.grid = Bunch.Bunch()
        for i in range(self.num_labels):
            self.grid[i] = aide.dc.Line(0, 0, 0, 0, color=aide.grid_fg,
                                        linestyle='dash', linewidth=1,
                                        alpha=self.grid_alpha,
                                        coord='window')
            self.objects.append(self.grid[i])

        self.axis_bg = aide.dc.Rectangle(0, 0, 100, 100, color=aide.norm_bg,
                                         fill=True, fillcolor=aide.axis_bg,
                                         coord='window')
        self.objects.append(self.axis_bg)

        self.lbls = Bunch.Bunch()
        for i in range(self.num_labels):
            self.lbls[i] = aide.dc.Text(0, 0, text='', color='black',
                                        font=self.font,
                                        fontsize=self.fontsize,
                                        coord='window')
            self.objects.append(self.lbls[i])

        self._title = aide.dc.Text(0, 0, text='', color='black',
                                   font=self.font, fontsize=self.fontsize,
                                   alpha=0.0,
                                   coord='window')
        self.objects.append(self._title)

    def update_elements(self, viewer):
        """This method is called if the plot is set with new points,
        or is scaled or panned with existing points.

        Update the XAxis labels to reflect the new values and/or pan/scale.
        """
        for i in range(self.num_labels):
            lbl = self.lbls[i]
            # get data coord equivalents
            x, y = self.get_data_xy(viewer, (lbl.x, lbl.y))
            # format according to user's preference
            lbl.text = self.format_value(x)

    def update_bbox(self, viewer, dims):
        """This method is called if the viewer's window is resized.

        Update all the XAxis elements to reflect the new dimensions.
        """
        title = self.get_title()
        self._title.text = title if title is not None else '555.55'
        self.title_wd, self.txt_ht = viewer.renderer.get_dimensions(self._title)

        wd, ht = dims[:2]
        y_hi = ht
        if title is not None:
            # remove Y space for X axis title
            y_hi -= self.txt_ht + 4
        # remove Y space for X axis labels
        y_hi -= self.txt_ht + self.pad_px

        self.aide.update_plot_bbox(y_hi=y_hi)

    def update_resize(self, viewer, dims, xy_lim):
        """This method is called if the viewer's window is resized.

        Update all the XAxis elements to reflect the new dimensions.
        """
        x_lo, y_lo, x_hi, y_hi = xy_lim
        wd, ht = dims[:2]

        # position axis title
        title = self.get_title()
        cx, cy = wd // 2 - self.title_wd // 2, ht - 4
        if title is not None:
            self._title.x = cx
            self._title.y = cy
            self._title.alpha = 1.0
            cy = cy - self.txt_ht
        else:
            self._title.alpha = 0.0

        # set X labels/grid as needed
        # calculate evenly spaced interval on X axis in window coords
        a = (x_hi - x_lo) // (self.num_labels - 1)
        cx = x_lo
        for i in range(self.num_labels):
            lbl = self.lbls[i]
            lbl.x, lbl.y = cx, cy
            # get data coord equivalents
            x, y = self.get_data_xy(viewer, (cx, cy))
            # convert to formatted label
            lbl.text = self.format_value(x)
            grid = self.grid[i]
            grid.x1 = grid.x2 = cx
            grid.y1, grid.y2 = y_lo, y_hi
            cx += a

        self.axis_bg.x1, self.axis_bg.x2 = 0, wd
        self.axis_bg.y1, self.axis_bg.y2 = y_hi, ht


class YAxis(Axis):
    """
    Plotable object that defines Y axis labels and grid lines.
    """
    def __init__(self, title=None, num_labels=4, font='sans', fontsize=10.0):
        super(YAxis, self).__init__(title=title, num_labels=num_labels,
                                    font=font, fontsize=fontsize)

        self.kind = 'axis_y'
        self.title_wd = 0
        self.txt_wd = 0
        self.txt_ht = 0
        self.pad_px = 4

    def register_decor(self, aide):
        self.aide = aide

        # add Y grid
        self.grid = Bunch.Bunch()
        for i in range(self.num_labels):
            self.grid[i] = aide.dc.Line(0, 0, 0, 0, color=aide.grid_fg,
                                        linestyle='dash', linewidth=1,
                                        alpha=self.grid_alpha,
                                        coord='window')
            self.objects.append(self.grid[i])

        # bg for RHS Y axis labels
        self.axis_bg = aide.dc.Rectangle(0, 0, 100, 100, color=aide.norm_bg,
                                         fill=True, fillcolor=aide.axis_bg,
                                         coord='window')
        self.objects.append(self.axis_bg)

        # bg for LHS Y axis title
        self.axis_bg2 = aide.dc.Rectangle(0, 0, 100, 100, color=aide.norm_bg,
                                          fill=True, fillcolor=aide.axis_bg,
                                          coord='window')
        self.objects.append(self.axis_bg2)

        # Y grid (tick) labels
        self.lbls = Bunch.Bunch()
        for i in range(self.num_labels):
            self.lbls[i] = aide.dc.Text(0, 0, text='', color='black',
                                        font=self.font,
                                        fontsize=self.fontsize,
                                        coord='window')
            self.objects.append(self.lbls[i])

        # Y title
        self._title = aide.dc.Text(0, 0, text=self.title, color='black',
                                   font=self.font,
                                   fontsize=self.fontsize,
                                   alpha=0.0,
                                   rot_deg=90.0,
                                   coord='window')
        self.objects.append(self._title)

    def update_elements(self, viewer):
        """This method is called if the plot is set with new points,
        or is scaled or panned with existing points.

        Update the YAxis labels to reflect the new values and/or pan/scale.
        """
        # set Y labels/grid as needed
        for i in range(self.num_labels):
            lbl = self.lbls[i]
            # get data coord equivalents
            x, y = self.get_data_xy(viewer, (lbl.x, lbl.y))
            lbl.text = self.format_value(y)

    def update_bbox(self, viewer, dims):
        """This method is called if the viewer's window is resized.

        Update all the YAxis elements to reflect the new dimensions.
        """
        title = self.get_title()
        self._title.text = title if title is not None else '555.55'

        wd, ht = dims[:2]
        self.title_wd, self.txt_ht = viewer.renderer.get_dimensions(self._title)
        # TODO: not sure this will give us the maximum length of number
        text = self.format_value(sys.float_info.max)
        t = self.aide.dc.Text(0, 0, text=text,
                              fontsize=self.fontsize, font=self.font)
        self.txt_wd, _ = viewer.renderer.get_dimensions(t)

        if title is not None:
            x_lo = self.txt_ht + 2 + self.pad_px
        else:
            x_lo = 0
        x_hi = wd - (self.txt_wd + 4) - self.pad_px

        self.aide.update_plot_bbox(x_lo=x_lo, x_hi=x_hi)

    def update_resize(self, viewer, dims, xy_lim):
        """This method is called if the viewer's window is resized.

        Update all the YAxis elements to reflect the new dimensions.
        """
        x_lo, y_lo, x_hi, y_hi = xy_lim
        wd, ht = dims[:2]

        # position axis title
        title = self.get_title()
        cx = self.txt_ht + 2
        cy = ht // 2 + self.title_wd // 2
        if title is not None:
            self._title.x = cx
            self._title.y = cy
            self._title.alpha = 1.0
        else:
            self._title.alpha = 0.0

        cx = x_hi + self.pad_px
        cy = y_hi
        # set Y labels/grid as needed
        a = (y_hi - y_lo) // (self.num_labels - 1)
        for i in range(self.num_labels):
            lbl = self.lbls[i]
            # calculate evenly spaced interval on Y axis in window coords
            lbl.x, lbl.y = cx, cy
            # get data coord equivalents
            x, y = self.get_data_xy(viewer, (cx, cy))
            lbl.text = self.format_value(y)
            grid = self.grid[i]
            grid.x1, grid.x2 = x_lo, x_hi
            grid.y1 = grid.y2 = cy
            cy -= a

        self.axis_bg.x1, self.axis_bg.x2 = x_hi, wd
        self.axis_bg.y1, self.axis_bg.y2 = y_lo, y_hi
        self.axis_bg2.x1, self.axis_bg2.x2 = 0, x_lo
        self.axis_bg2.y1, self.axis_bg2.y2 = y_lo, y_hi


class PlotBG(CompoundObject):
    """
    Plotable object that defines the plot background.

    Can include a warning line and an alert line.  If the last Y value
    plotted exceeds the warning line then the background changes color.
    For example, you might be plotting detector values and want to set
    a warning if a certain threshold is crossed and an alert if the
    detector has saturated (alerts are higher than warnings).
    """
    def __init__(self, warn_y=None, alert_y=None, linewidth=1):
        super(PlotBG, self).__init__()

        self.y_lbl_info = [warn_y, alert_y]
        self.warn_y = warn_y
        self.alert_y = alert_y
        self.linewidth = linewidth
        # default warning check
        self.check_warning = self._check_warning

        self.norm_bg = 'white'
        self.warn_bg = 'lightyellow'
        self.alert_bg = 'mistyrose2'
        self.kind = 'plot_bg'
        self.pickable = True
        self.opaque = True

    def register_decor(self, aide):
        self.aide = aide

        # add a backdrop that we can change color for visual warnings
        self.bg = aide.dc.Rectangle(0, 0, 100, 100, color=aide.norm_bg,
                                    fill=True, fillcolor=aide.norm_bg,
                                    fillalpha=1.0,
                                    coord='window')
        self.objects.append(self.bg)

        # add warning and alert lines
        self.ln_warn = aide.dc.Line(0, self.warn_y, 1, self.warn_y,
                                    color='gold3', linewidth=self.linewidth,
                                    alpha=0.0, coord='window')
        self.objects.append(self.ln_warn)

        self.ln_alert = aide.dc.Line(0, self.alert_y, 1, self.alert_y,
                                     color='red', linewidth=self.linewidth,
                                     alpha=0.0, coord='window')
        self.objects.append(self.ln_alert)

    def warning(self):
        self.bg.fillcolor = self.warn_bg

    def alert(self):
        self.bg.fillcolor = self.alert_bg

    def normal(self):
        self.bg.fillcolor = self.norm_bg

    def _check_warning(self):
        max_y = None
        for i, plot_src in enumerate(self.aide.plots.values()):
            limits = plot_src.get_limits('data')
            y = limits[1][1]
            max_y = y if max_y is None else max(max_y, y)

        if max_y is not None:
            if self.alert_y is not None and max_y > self.alert_y:
                self.alert()
            elif self.warn_y is not None and max_y > self.warn_y:
                self.warning()
            else:
                self.normal()

    def update_elements(self, viewer):
        """This method is called if the plot is set with new points,
        or is scaled or panned with existing points.

        Update the XAxis labels to reflect the new values and/or pan/scale.
        """
        y_lo, y_hi = self.aide.bbox.T[1].min(), self.aide.bbox.T[1].max()
        # adjust warning/alert lines
        if self.warn_y is not None:
            x, y = self.get_canvas_xy(viewer, (0, self.warn_y))
            if y_lo <= y <= y_hi:
                self.ln_warn.alpha = 1.0
            else:
                # y out of range of plot area, so make it invisible
                self.ln_warn.alpha = 0.0

            self.ln_warn.y1 = self.ln_warn.y2 = y

        if self.alert_y is not None:
            x, y = self.get_canvas_xy(viewer, (0, self.alert_y))
            if y_lo <= y <= y_hi:
                self.ln_alert.alpha = 1.0
            else:
                # y out of range of plot area, so make it invisible
                self.ln_alert.alpha = 0.0

            self.ln_alert.y1 = self.ln_alert.y2 = y

        self.check_warning()

    def update_bbox(self, viewer, dims):
        # this object does not adjust the plot bbox at all
        pass

    def update_resize(self, viewer, dims, xy_lim):
        """This method is called if the viewer's window is resized.

        Update all the PlotBG elements to reflect the new dimensions.
        """
        # adjust bg to window size, in case it changed
        x_lo, y_lo, x_hi, y_hi = xy_lim
        wd, ht = dims[:2]

        self.bg.x1, self.bg.y1 = x_lo, y_lo
        self.bg.x2, self.bg.y2 = x_hi, y_hi

        # adjust warning/alert lines
        if self.warn_y is not None:
            x, y = self.get_canvas_xy(viewer, (0, self.warn_y))
            self.ln_warn.x1, self.ln_warn.x2 = x_lo, x_hi
            self.ln_warn.y1 = self.ln_warn.y2 = y

        if self.alert_y is not None:
            x, y = self.get_canvas_xy(viewer, (0, self.alert_y))
            self.ln_alert.x1, self.ln_alert.x2 = x_lo, x_hi
            self.ln_alert.y1 = self.ln_alert.y2 = y

    def add_plot(self, viewer, plot_src):
        pass

    def delete_plot(self, viewer, plot_src):
        pass

    def get_canvas_xy(self, viewer, pt):
        arr_pts = np.asarray(pt)
        return viewer.tform['data_to_plot'].to_(arr_pts).T[:2]


class PlotTitle(CompoundObject):
    """
    Plotable object that defines the plot title and keys.
    """
    def __init__(self, title='', font='sans', fontsize=12.0):
        super(PlotTitle, self).__init__()

        self.font = font
        self.fontsize = fontsize
        self.title = title
        self.txt_ht = 0
        self.kind = 'plot_title'
        self.format_label = self._format_label
        self.pad_px = 5

    def register_decor(self, aide):
        self.aide = aide

        self.title_bg = aide.dc.Rectangle(0, 0, 100, 100, color=aide.norm_bg,
                                          fill=True, fillcolor=aide.axis_bg,
                                          coord='window')
        self.objects.append(self.title_bg)

        self.lbls = dict()
        self.lbls[0] = aide.dc.Text(0, 0, text=self.title, color='black',
                                    font=self.font,
                                    fontsize=self.fontsize,
                                    coord='window')
        self.objects.append(self.lbls[0])

    def _format_label(self, lbl, plot_src):
        """Default formatter for PlotTitle labels.
        """
        lbl.text = "{0:}".format(plot_src.name)

    def update_elements(self, viewer):
        """This method is called if the plot is set with new points,
        or is scaled or panned with existing points.

        Update the PlotTitle labels to reflect the new values.
        """
        for i, plot_src in enumerate(self.aide.plots.values()):
            lbl = self.lbls[plot_src]
            self.format_label(lbl, plot_src)

    def update_bbox(self, viewer, dims):
        """This method is called if the viewer's window is resized.

        Update all the PlotTitle elements to reflect the new dimensions.
        """
        wd, ht = dims[:2]
        if self.txt_ht == 0:
            _, self.txt_ht = viewer.renderer.get_dimensions(self.lbls[0])

        y_lo = self.txt_ht + self.pad_px

        self.aide.update_plot_bbox(y_lo=y_lo)

    def update_resize(self, viewer, dims, xy_lim):
        """This method is called if the viewer's window is resized.

        Update all the PlotTitle elements to reflect the new dimensions.
        """
        x_lo, y_lo, x_hi, y_hi = xy_lim
        wd, ht = dims[:2]

        nplots = len(list(self.aide.plots.keys())) + 1

        # set title labels as needed
        a = wd // (nplots + 1)

        cx, cy = 4, self.txt_ht
        lbl = self.lbls[0]
        lbl.x, lbl.y = cx, cy

        for i, plot_src in enumerate(self.aide.plots.values()):
            cx += a
            lbl = self.lbls[plot_src]
            lbl.x, lbl.y = cx, cy
            self.format_label(lbl, plot_src)

        self.title_bg.x1, self.title_bg.x2 = 0, wd
        self.title_bg.y1, self.title_bg.y2 = 0, y_lo

    def add_plot(self, viewer, plot_src):
        text = plot_src.name
        color = plot_src.color
        lbl = self.aide.dc.Text(0, 0, text=text, color=color,
                                font=self.font,
                                fontsize=self.fontsize,
                                coord='window')
        self.lbls[plot_src] = lbl
        self.objects.append(lbl)
        lbl.crdmap = self.lbls[0].crdmap
        self.format_label(lbl, plot_src)

        # reorder and place labels
        dims = viewer.get_window_size()
        self.update_resize(viewer, dims, self.aide.llur)

    def delete_plot(self, viewer, plot_src):
        lbl = self.lbls[plot_src]
        del self.lbls[plot_src]
        self.objects.remove(lbl)

        # reorder and place labels
        dims = viewer.get_window_size()
        self.update_resize(viewer, dims, self.aide.llur)


class CalcPlot(XYPlot):

    def __init__(self, name=None, x_fn=None, y_fn=None, color='black',
                 linewidth=1, linestyle='solid', alpha=1.0, **kwdargs):
        super(CalcPlot, self).__init__(name=name,
                                       color=color, linewidth=linewidth,
                                       linestyle=linestyle, alpha=alpha,
                                       **kwdargs)
        self.kind = 'calcplot'
        if x_fn is None:
            x_fn = lambda x: x   # noqa
        self.x_fn = x_fn
        if y_fn is None:
            y_fn = lambda y: y   # noqa
        self.y_fn = y_fn

    def plot(self, y_fn, x_fn=None):
        if x_fn is not None:
            self.x_fn = x_fn
        self.y_fn = y_fn
        self.plot_xlim = (None, None)

    def calc_points(self, viewer, start_x, stop_x):
        # in case X axis is flipped
        start_x, stop_x = min(start_x, stop_x), max(start_x, stop_x)

        new_xlim = (start_x, stop_x)
        if new_xlim == self.plot_xlim:
            # X limits are the same, no need to recalculate points
            return

        self.plot_xlim = new_xlim

        wd, ht = self.viewer.get_window_size()

        x_pts = self.x_fn(np.linspace(start_x, stop_x, wd, dtype=float))
        y_pts = self.y_fn(x_pts)
        points = np.array((x_pts, y_pts)).T
        self.path.points = points

    def get_limits(self, lim_type):
        try:
            llur = self.path.get_llur()
            limits = [llur[0:2], llur[2:4]]
            return np.array(limits)
        except Exception:
            return np.array(((0.0, 0.0), (0.0, 0.0)))


# register our types
register_canvas_types(dict(xyplot=XYPlot, calcplot=CalcPlot))



#
# time_series.py -- utilities to assist in making time-series plots
#
import time

from ginga.canvas.types.plots import XAxis, PlotTitle, PlotBG


class XTimeAxis(XAxis):
    """Override basic XAxis to print X values as times.
    """

    def _format_value(self, t):
        return time.strftime("%H:%M:%S", time.localtime(t))


class TimePlotTitle(PlotTitle):
    """Override basic PlotTitle to print current value of latest data
    points.
    """

    def _format_label(self, lbl, plot_src):
        pt = plot_src.get_latest()
        if pt is None:
            lbl.text = "{0:}".format(plot_src.name)
        else:
            x, y = pt[:2]
            lbl.text = "{0:}: {1: .2f}".format(plot_src.name, y)


class TimePlotBG(PlotBG):
    """Override basic PlotBg to make warnings and alerts happen if the
    most recent value exceeds the thresholds.
    """

    def _check_warning(self):
        max_y = None
        for i, plot_src in enumerate(self.aide.plots.values()):
            pt = plot_src.get_latest()
            if pt is not None:
                x, y = pt[:2]
                max_y = y if max_y is None else max(max_y, y)

        if max_y is not None:
            if self.alert_y is not None and max_y > self.alert_y:
                self.alert()
            elif self.warn_y is not None and max_y > self.warn_y:
                self.warning()
            else:
                self.normal()



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os.path

from ginga.util import loader
from ginga.gw import Widgets

from .base import Stage, StageError


class Input(Stage):

    _stagename = 'input'

    def __init__(self):
        super().__init__()

        self.image = None
        self._path = ""

    def build_gui(self, container):
        fr = Widgets.Frame("Input")

        vbox = Widgets.VBox()
        self.w.lbl = Widgets.TextEntrySet('', editable=True)
        self.w.lbl.set_text(self._path)
        self.w.lbl.set_tooltip("Enter a path to load the data")
        vbox.add_widget(self.w.lbl, stretch=0)
        self.w.lbl.add_callback('activated', self.set_path_cb)

        fr.set_widget(vbox)
        container.set_widget(fr)

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, val):
        self._path = val
        if self.gui_up:
            self.w.lbl.set_text(val)

    def _set_image(self, image):
        self.image = image
        self.path = image.get('path', "")
        _dir, _fn = None, None
        if len(self._path) > 0:
            _dir, _fn = os.path.split(self._path)
        self.pipeline.set(input_image=image,
                          input_folder=_dir, input_filename=_fn)

    def set_image(self, image):
        self._set_image(image)
        self.pipeline.run_from(self)

    def _set_path(self, path):
        image = loader.load_data(path, logger=self.logger)
        self._set_image(image)

    def set_path_cb(self, widget):
        path = widget.get_text().strip()
        self._set_path(path)
        self.pipeline.run_from(self)

    def run(self, prev_stage):
        assert prev_stage is None, StageError("'input' in wrong location")

        if self._bypass:
            self.pipeline.send(res_np=None)
            return

        data_np = self.image.get_data()
        self.pipeline.send(res_np=data_np)

    def _get_state(self):
        return dict(path=self._path)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)

        self.path = d['path']



#
# compat.py -- Python compatibility shims.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
if sys.version_info < (3, 8):
    # Python 3.7
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


def ep_get(group):
    """Get the entry points for a group."""
    ep = entry_points()
    if hasattr(ep, "select"):
        return ep.select(group=group)
    else:
        return ep.get(group, [])



# Figure out which widget set we are using and import those wrappers
from ginga import toolkit
tkname = toolkit.get_family()

if tkname == 'qt':
    from ginga.qtw.Plot import *  # noqa

elif tkname == 'gtk3':
    from ginga.gtk3w.Plot import *  # noqa

elif tkname == 'gtk4':
    from ginga.gtk4w.Plot import *  # noqa

elif tkname == 'pg':
    from ginga.web.pgw.Plot import *  # noqa

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
import os.path
import glob
import importlib
import inspect

from . import base


def get_stage_catalog(logger):
    """Return a dictionary containing all the stages that could be loaded.
    """
    stage_catalog = dict()
    thisdir, _ = os.path.split(base.__file__)
    for pypath in glob.glob(os.path.join(thisdir, "*.py")):
        _, pyfile = os.path.split(pypath)
        modname, _ext = os.path.splitext(pyfile)
        try:
            mod = importlib.import_module('ginga.util.stages.' + modname)

        except ImportError as e:
            logger.error("Couldn't import '{}': {}".format(modname, e))
            continue

        for attr_name in dir(mod):
            attr = getattr(mod, attr_name)
            if (inspect.isclass(attr) and issubclass(attr, base.Stage) and
                attr is not base.Stage):
                name = getattr(attr, '_stagename', str(attr).lower())
                if name.startswith('viewer-') or name.startswith('rgbmap-'):
                    continue
                # for now, do not include viewer
                stage_catalog[name] = attr

    return stage_catalog
